"""
Copyright 2008-2015 Free Software Foundation, Inc.
This file is part of GNU Radio

GNU Radio Companion is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

GNU Radio Companion is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""

from __future__ import absolute_import

import collections
import itertools
import copy

import re

import ast

from ._flags import Flags

from ..base import Element
from ..utils.descriptors import lazy_property

def _get_elem(iterable, key):
    items = list(iterable)
    for item in items:
        if item.key == key:
            return item
    return ValueError('Key "{}" not found in {}.'.format(key, items))

class Block(Element):

    is_block = True

    key = ''
    label = ''
    category = ''
    flags = Flags('')
    documentation = {'': ''}

    value = None
    asserts = []

    parameters_data = []
    inputs_data = []
    outputs_data = []

    e_left_data = []
    e_right_data = []
    e_top_data = []
    e_bottom_data = []

    b_left_data = []
    b_right_data = []
    b_top_data = []
    b_bottom_data = []

    extra_data = {}
    loaded_from = '(unknown)'

    def __init__(self, parent):
        """Make a new block from nested data."""
        super(Block, self).__init__(parent)
        param_factory = self.parent_platform.make_param
        port_factory = self.parent_platform.make_port

        self.params = collections.OrderedDict(
            (data['id'], param_factory(parent=self, **data))
            for data in self.parameters_data
        )
        if self.key == 'options':
            self.params['id'].hide = 'part'

        self.sinks = [port_factory(parent=self, **params) for params in self.inputs_data]
        self.sources = [port_factory(parent=self, **params) for params in self.outputs_data]

        self.e_lefts = [port_factory(parent=self, **params) for params in self.e_left_data]
        self.e_rights = [port_factory(parent=self, **params) for params in self.e_right_data]
        self.e_tops = [port_factory(parent=self, **params) for params in self.e_top_data]
        self.e_bottoms = [port_factory(parent=self, **params) for params in self.e_bottom_data]

        self.b_lefts = [port_factory(parent=self, **params) for params in self.b_left_data]
        self.b_rights = [port_factory(parent=self, **params) for params in self.b_right_data]
        self.b_tops = [port_factory(parent=self, **params) for params in self.b_top_data]
        self.b_bottoms = [port_factory(parent=self, **params) for params in self.b_bottom_data]

        self.active_sources = []  # on rewrite
        self.active_sinks = []  # on rewrite

        self.states = {'state': True}

    # region Rewrite_and_Validation
    def rewrite(self):
        """
        Add and remove ports to adjust for the nports.
        """
        Element.rewrite(self)

        # Adjust nports

        t_ports = (
          self.sources, self.sinks,
          self.e_lefts, self.e_rights, self.e_tops, self.e_bottoms,
          self.b_lefts, self.b_rights, self.b_tops, self.b_bottoms,
        )

        for ports in t_ports:
            self._rewrite_nports(ports)

        # disconnect hidden ports

        self.active_sources = [p for p in self.sources]
        self.active_sinks = [p for p in self.sinks]

#       e nodes: don't bother with "active" e nodes since we don't have active
#       and hidden in gseim -> use e_lefts rather than active_e_lefts (etc)

    def _rewrite_nports(self, ports):
        for port in ports:
            if hasattr(port, 'master_port'):  # Not a master port and no left-over clones
                port.dtype = port.master_port.dtype
                continue

    def validate(self):
        """
        Validate this block.
        Call the base class validate.
        Evaluate the checks: each check must evaluate to True.
        """
        Element.validate(self)
        self._run_asserts()
        self._validate_var_value()

    def _run_asserts(self):
        """Evaluate the checks"""
        for expr in self.asserts:
            try:
                if not self.evaluate(expr):
                    self.add_error_message('Assertion "{}" failed.'.format(expr))
            except Exception:
                self.add_error_message('Assertion "{}" did not evaluate.'.format(expr))

    def _validate_var_value(self):
        """or variables check the value (only if var_value is used)"""
        if self.is_variable and self.value != 'value':
            try:
                self.parent_flowgraph.evaluate(self.value, local_namespace=self.namespace)
            except Exception as err:
                self.add_error_message('Value "{}" cannot be evaluated:\n{}'.format(self.value, err))
    # endregion

    # region Properties

    def __str__(self):
        return 'Block - {} - {}({})'.format(self.name, self.label, self.key)

    def __repr__(self):
        try:
            name = self.name
        except Exception:
            name = self.key
        return 'block[' + name + ']'

    @property
    def name(self):
        return self.params['id'].value

    @lazy_property
    def is_variable(self):
        return bool(self.value)

    @lazy_property
    def is_import(self):
        return self.key == 'import'

    @lazy_property
    def is_snippet(self):
        return self.key == 'snippet'

    @property
    def comment(self):
        return self.params['comment'].value

    @property
    def state(self):
        return 'enabled'

    @state.setter
    def state(self, value):
        """Sets the state for the block."""
        self.states['state'] = value

    # Enable/Disable Aliases
    @property
    def enabled(self):
        """Get the enabled state of the block"""
        return self.state != 'disabled'

    def ports(self):
        return itertools.chain(self.sources, self.sinks,
           self.e_lefts, self.e_rights, self.e_tops, self.e_bottoms,
           self.b_lefts, self.b_rights, self.b_tops, self.b_bottoms)

    def active_ports(self):
        return itertools.chain(
            self.active_sources,
            self.active_sinks,
            self.e_lefts,
            self.e_rights,
            self.e_tops,
            self.e_bottoms,
            self.b_lefts,
            self.b_rights,
            self.b_tops,
            self.b_bottoms,
        )

    def children(self):
        return itertools.chain(self.params.values(), self.ports())

    # Access

    def get_sink(self, key):
        return _get_elem(self.sinks, key)

    def get_source(self, key):
        return _get_elem(self.sources, key)

    def get_e_left(self, key):
        return _get_elem(self.e_lefts, key)

    def get_e_right(self, key):
        return _get_elem(self.e_rights, key)

    def get_e_top(self, key):
        return _get_elem(self.e_tops, key)

    def get_e_bottom(self, key):
        return _get_elem(self.e_bottoms, key)

    def get_b_left(self, key):
        return _get_elem(self.b_lefts, key)

    def get_b_right(self, key):
        return _get_elem(self.b_rights, key)

    def get_b_top(self, key):
        return _get_elem(self.b_tops, key)

    def get_b_bottom(self, key):
        return _get_elem(self.b_bottoms, key)

    # Resolve
    @property
    def namespace(self):
        return {key: param.get_evaluated() for key, param in iter(self.params.items())}

    def evaluate(self, expr):
        return self.parent_flowgraph.evaluate(expr, self.namespace)

    # Import/Export Methods
    def export_data(self):
        """
        Export this block's params to nested data.

        Returns:
            a nested data odict
        """
        data = collections.OrderedDict()
        if self.key != 'options':
            data['name'] = self.name
            data['id'] = self.key
        data['parameters'] = collections.OrderedDict(sorted(
            (param_id, param.value) for param_id, param in self.params.items()
            if (param_id != 'id' or self.key == 'options')
        ))
        data['states'] = collections.OrderedDict(sorted(self.states.items()))

        return data

    def import_data(self, name, states, parameters, **_):
        """
        Import this block's params from nested data.
        Any param keys that do not exist will be ignored.
        Since params can be dynamically created based another param,
        call rewrite, and repeat the load until the params stick.
        """
        self.params['id'].value = name
        self.states.update(states)

        def get_hash():
            return hash(tuple(hash(v) for v in self.params.values()))

        pre_rewrite_hash = -1
        while pre_rewrite_hash != get_hash():
            for key, value in iter(parameters.items()):
                try:
                    self.params[key].set_value(value)
                except KeyError:
                    continue
            # Store hash and call rewrite
            pre_rewrite_hash = get_hash()
            self.rewrite()
            
    # Controller Modify




import collections
import os

import codecs

from .top_block import TopBlockGenerator

from .. import Constants
from ..io import yaml

class HierBlockGenerator(TopBlockGenerator):
    """Extends the top block generator to also generate a block YML file"""

    def __init__(self, flow_graph, file_path):
        """
        Initialize the hier block generator object.

        Args:
            flow_graph: the flow graph object
            file_path: where to write the py file (the yml goes into HIER_BLOCK_LIB_DIR)
        """
        TopBlockGenerator.__init__(self, flow_graph, file_path)
        platform = flow_graph.parent

        hier_block_lib_dir = platform.config.hier_block_lib_dir
        if not os.path.exists(hier_block_lib_dir):
            os.mkdir(hier_block_lib_dir)

        self._mode = Constants.HIER_BLOCK_FILE_MODE
        print('HierBlockGenerator: hier_block_lib_dir:', hier_block_lib_dir)
        self.file_path = os.path.join(hier_block_lib_dir, self._flow_graph.get_option('id'))
        self.file_path_yml = self.file_path + '.hblock.yml'
        print('HierBlockGenerator: self.file_path_yml:', self.file_path_yml)

    def write(self, flow_graph):
        """generate output and write it to files"""
        print('HierBlockGenerator: calling TopBlockGenerator.write')
        TopBlockGenerator.write(self)
        print('HierBlockGenerator: TopBlockGenerator.write over')

        data = yaml.dump(self._build_block_n_from_flow_graph_io(flow_graph))

        replace = [
            ('parameters:', '\nparameters:'),
            ('inputs:', '\ninputs:'),
            ('outputs:', '\noutputs:'),
            ('e_left_nodes:', '\ne_left_nodes:'),
            ('e_right_nodes:', '\ne_right_nodes:'),
            ('e_top_nodes:', '\ne_top_nodes:'),
            ('e_bottom_nodes:', '\ne_bottom_nodes:'),
            ('b_left_nodes:', '\nb_left_nodes:'),
            ('b_right_nodes:', '\nb_right_nodes:'),
            ('b_top_nodes:', '\nb_top_nodes:'),
            ('b_bottom_nodes:', '\nb_bottom_nodes:'),
            ('asserts:', '\nasserts:'),
            ('templates:', '\ntemplates:'),
            ('documentation:', '\ndocumentation:'),
        ]

        for r in replace:
            data = data.replace(*r)

        print('HierBlockGenerator: self.file_path_yml:', self.file_path_yml)
        with codecs.open(os.path.expanduser(self.file_path_yml), 'w', encoding='utf-8') as fp:
            fp.write(data)

        # Windows only supports S_IREAD and S_IWRITE, other flags are ignored
        os.chmod(self.file_path_yml, self._mode)

    def _build_block_n_from_flow_graph_io(self, flow_graph):
        """
        Generate a block YML nested data from the flow graph IO

        Returns:
            a yml node tree
        """
        # Extract info from the flow graph
        block_id = self._flow_graph.get_option('id')
        parameters = self._flow_graph.get_parameters()

        def var_or_value(name):
            if name in (p.name for p in parameters):
                return "${" + name + " }"
            return name

        # Build the nested data
        data = collections.OrderedDict()
        data['id'] = block_id
        data['label'] = (
            self._flow_graph.get_option('title') or
            self._flow_graph.get_option('id').replace('_', ' ').title()
        )
        data['category'] = self._flow_graph.get_option('category')

#       Parameters
        data['parameters'] = []

        if 'name' not in [dx['id'] for dx in data['parameters']]:
            p = collections.OrderedDict()
            p['id'] = 'name'
            p['label'] = 'name'
            p['dtype'] = 'string'
            p['default'] = 'none'
            data['parameters'].append(p)

        if 'rotate_strict' not in flow_graph.gparms.keys():
            flow_graph.gparms['rotate_strict'] = 'no'

        if 'mirror' not in flow_graph.gparms.keys():
            flow_graph.gparms['mirror'] = 'none'

        if 'port_sep_x' not in flow_graph.gparms.keys():
            flow_graph.gparms['port_sep_x'] = '4'
        if 'port_sep_y' not in flow_graph.gparms.keys():
            flow_graph.gparms['port_sep_y'] = '4'
        if 'port_block_x' not in flow_graph.gparms.keys():
            flow_graph.gparms['port_block_x'] = '4'
        if 'port_block_y' not in flow_graph.gparms.keys():
            flow_graph.gparms['port_block_y'] = '4'

        if 'drawing_scheme' in flow_graph.gparms.keys():
            if flow_graph.gparms['drawing_scheme'] == 'symbol':
                if 'port_offset_l' not in flow_graph.gparms.keys():
                    flow_graph.gparms['port_offset_l'] = '0'
                if 'port_offset_r' not in flow_graph.gparms.keys():
                    flow_graph.gparms['port_offset_r'] = '0'
                if 'port_offset_t' not in flow_graph.gparms.keys():
                    flow_graph.gparms['port_offset_t'] = '0'
                if 'port_offset_b' not in flow_graph.gparms.keys():
                    flow_graph.gparms['port_offset_b'] = '0'
        else:
            flow_graph.gparms['drawing_scheme'] = 'name'

#       write gparms which the user would be interested in editing (when the sub ckt is
#       called), then followed by gparms such as port_sep_x which the user would not
#       be editing.

        l1 = [
          'rotate_strict',
          'mirror',
          'port_sep_x',
          'port_sep_y',
          'port_block_x',
          'port_block_y',
          'drawing_scheme',
          'port_offset_l',
          'port_offset_r',
          'port_offset_t',
          'port_offset_b',
        ]

        for k, v in flow_graph.gparms.items():
            if k not in l1:
                p = collections.OrderedDict()
                p['id'] = k
                p['label'] = k
                p['default'] = v
                data['parameters'].append(p)

        for k, v in flow_graph.gparms.items():
            if k in l1:
                p = collections.OrderedDict()
                p['id'] = k
                p['label'] = k
                p['default'] = v
                data['parameters'].append(p)

#       outvars
        data['outvars'] = []

        for k in flow_graph.outvars.keys():
            data['outvars'].append(k)

        # Ports
        t_dir = (
          'inputs', 'outputs',
          'e_left_nodes', 'e_right_nodes', 'e_top_nodes', 'e_bottom_nodes',
          'b_left_nodes', 'b_right_nodes', 'b_top_nodes', 'b_bottom_nodes',
        )
        for direction in t_dir:
            data[direction] = []
            for port in get_hier_block_io(self._flow_graph, direction):
                p = collections.OrderedDict()
                p['label'] = port.parent.params['label'].value
                p['dtype'] = port.dtype

                print('HierBlockGenerator: _build_block_n_from_flow_graph_io: p[label]:', p['label'])

                data[direction].append(p)

        data['grc_source'] = str(self._flow_graph.grc_file_path)

        return data

def get_hier_block_io(flow_graph, direction, domain=None):
    """
    Get a list of io ports for this flow graph.

    Returns a list of blocks
    """

    if direction == 'inputs':
        pads = flow_graph.get_pad_sources()
    elif direction == 'outputs':
        pads = flow_graph.get_pad_sinks()
    elif direction == 'e_left_nodes':
        pads = flow_graph.get_pad_e_lefts()
    elif direction == 'e_right_nodes':
        pads = flow_graph.get_pad_e_rights()
    elif direction == 'e_top_nodes':
        pads = flow_graph.get_pad_e_tops()
    elif direction == 'e_bottom_nodes':
        pads = flow_graph.get_pad_e_bottoms()
    elif direction == 'b_left_nodes':
        pads = flow_graph.get_pad_b_lefts()
    elif direction == 'b_right_nodes':
        pads = flow_graph.get_pad_b_rights()
    elif direction == 'b_top_nodes':
        pads = flow_graph.get_pad_b_tops()
    elif direction == 'b_bottom_nodes':
        pads = flow_graph.get_pad_b_bottoms()

    for pad in pads:
        if direction == 'inputs':
            for port in pad.sources:
                yield port
        elif direction == 'outputs':
            for port in pad.sinks:
                yield port
        elif direction == 'e_left_nodes':
            for port in pad.e_rights:
                yield port
        elif direction == 'e_right_nodes':
            for port in pad.e_lefts:
                yield port
        elif direction == 'e_top_nodes':
            for port in pad.e_bottoms:
                yield port
        elif direction == 'e_bottom_nodes':
            for port in pad.e_tops:
                yield port
        elif direction == 'b_left_nodes':
            for port in pad.b_rights:
                yield port
        elif direction == 'b_right_nodes':
            for port in pad.b_lefts:
                yield port
        elif direction == 'b_top_nodes':
            for port in pad.b_bottoms:
                yield port
        elif direction == 'b_bottom_nodes':
            for port in pad.b_tops:
                yield port



"""
Copyright (C) 2022 - Mahesh Patil <mbpatil@ee.iitb.ac.in>
This file is part of GSEIM.

GSEIM is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""


import gi
import os
import time
from subprocess import Popen, PIPE, STDOUT
from gi.repository import GLib
from ..core.utils import gutils as gu

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, Pango

class RunProcess(Gtk.Dialog):
    def __init__(self, parent, cmd, s_max_time):
        Gtk.Dialog.__init__(self, title="Run Process", transient_for=parent, flags=0)

        self.add_button("Stop program", Gtk.ResponseType.CANCEL)
        self.add_button("Close window", Gtk.ResponseType.OK)

        self.set_default_size(600, 600)
        self.set_position(Gtk.WindowPosition.CENTER)

        grid = Gtk.Grid()
        grid.set_row_spacing(10)
        grid.set_column_spacing(10)

        self.s1 = ''
        self.label1 = Gtk.Label(label = self.s1, xalign=0)
        self.label1.modify_font(Pango.font_description_from_string('DejaVu Sans Mono 9'))

        grid.add(self.label1)

        self.scrolled_window = Gtk.ScrolledWindow()
        self.scrolled_window.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.ALWAYS)
        self.scrolled_window.show()
        self.vbox.pack_start(self.scrolled_window, True, True, 0)

        self.scrolled_window.add(grid)

        self.t_end = time.time() + gu.get_sec(s_max_time)
        self.t_1 = 0.0
        self.flag_normal = False
        self.flag_error = False
        self.flag_timelimit = False
        self.flag_killed = False
        self.sub_proc = Popen(cmd, stdout=PIPE, shell=True)

#       first argument is in milliseconds
        GLib.timeout_add(100, self.update_terminal)

        self.show_all()

    def update_terminal(self):
        NMAX = 100
        self.s1 = self.label1.get_text() + gu.non_block_read(self.sub_proc.stdout)
        n_lines = self.s1.count('\n')
        m = n_lines - NMAX
        if m > 0:
            n1 = gu.find_nth(self.s1, '\n', m)
            self.s1 = self.s1[n1 + 1:]

        self.label1.set_text(self.s1)

        position = self.scrolled_window.get_vadjustment()
        position.set_value(position.get_upper() - position.get_page_size())
        self.scrolled_window.set_vadjustment(position)

        self.t_1 = time.time()

        if self.t_1 > self.t_end:
            print('update_terminal: time exceeded, killing process')

            while self.sub_proc.poll() is None:
                os.kill(self.sub_proc.pid, 9)

            self.flag_timelimit = True
            return False

        if not self.sub_proc.poll() is None:
            if 'program completed' in self.s1.lower():
                print('update_terminal: program completed found.')
                self.flag_normal = True
            else:
                print('update_terminal: program completed not found.')
                self.flag_error = True
            return False

        return True



# Copyright 2008-2016 Free Software Foundation, Inc.
# This file is part of GNU Radio
#
# GNU Radio Companion is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# GNU Radio Companion is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

from __future__ import absolute_import

from .. import Constants
from ..base import Element
from ..utils.descriptors import (
    EvaluatedFlag, EvaluatedEnum, EvaluatedPInt,
    setup_names, lazy_property
)

@setup_names
class Port(Element):

    is_port = True

    dtype = EvaluatedEnum(list(Constants.TYPE_TO_SIZEOF.keys()), default='')

    def __init__(self, parent, direction, id, label='', domain=Constants.DEFAULT_DOMAIN, dtype='',
                 bus_struct=None, **_):
        """Make a new port from nested data."""
        Element.__init__(self, parent)

        self._dir = direction

        self.port_subtype = direction if direction != 'source' else ''

        if direction in ('source', 'sink'):
            self.port_type = 'flowgraph'
        elif direction in ('e_left', 'e_right', 'e_top', 'e_bottom'):
            self.port_type = 'electrical'
        elif direction in ('b_left', 'b_right', 'b_top', 'b_bottom'):
            self.port_type = 'bus'

        self.key = id
        if not label:

            label = id if not id.isdigit() else {
                'sink': 'in',
                'source': 'out',
                'e_left': 'e',
                'e_right': 'e',
                'e_top': 'e',
                'e_bottom': 'e',
                'b_left': 'b',
                'b_right': 'b',
                'b_top': 'b',
                'b_bottom': 'b',
            }[direction]

        self.name = self._base_name = label

        self.domain = domain
        self.dtype = dtype

        self.stored_hidden_state = None
        self.bus_structure = bus_struct

        # end of args ########################################################

    def __str__(self):
        if self.is_source:
            return 'Source - {}({})'.format(self.name, self.key)
        if self.is_sink:
            return 'Sink - {}({})'.format(self.name, self.key)
        if self.is_e_left:
            return 'e_left - {}({})'.format(self.name, self.key)
        if self.is_e_right:
            return 'e_right - {}({})'.format(self.name, self.key)
        if self.is_e_top:
            return 'e_top - {}({})'.format(self.name, self.key)
        if self.is_e_bottom:
            return 'e_bottom - {}({})'.format(self.name, self.key)
        if self.is_b_left:
            return 'b_left - {}({})'.format(self.name, self.key)
        if self.is_b_right:
            return 'b_right - {}({})'.format(self.name, self.key)
        if self.is_b_top:
            return 'b_top - {}({})'.format(self.name, self.key)
        if self.is_b_bottom:
            return 'b_bottom - {}({})'.format(self.name, self.key)

    def __repr__(self):
        if self.is_source:
            s1 = 'sources'
        elif self.is_sink:
            s1 = 'sinks'
        elif self.is_e_left:
            s1 = 'e_lefts'
        elif self.is_e_right:
            s1 = 'e_rights'
        elif self.is_e_top:
            s1 = 'e_tops'
        elif self.is_e_bottom:
            s1 = 'e_bottoms'
        elif self.is_b_left:
            s1 = 'b_lefts'
        elif self.is_b_right:
            s1 = 'b_rights'
        elif self.is_b_top:
            s1 = 'b_tops'
        elif self.is_b_bottom:
            s1 = 'b_bottoms'

        return '{!r}.{}[{}]'.format(self.parent, s1, self.key)

    @lazy_property
    def is_sink(self):
        return self._dir == 'sink'

    @lazy_property
    def is_source(self):
        return self._dir == 'source'

    @lazy_property
    def is_e_left(self):
        return self._dir == 'e_left'

    @lazy_property
    def is_e_right(self):
        return self._dir == 'e_right'

    @lazy_property
    def is_e_top(self):
        return self._dir == 'e_top'

    @lazy_property
    def is_e_bottom(self):
        return self._dir == 'e_bottom'

    @lazy_property
    def is_b_left(self):
        return self._dir == 'b_left'

    @lazy_property
    def is_b_right(self):
        return self._dir == 'b_right'

    @lazy_property
    def is_b_top(self):
        return self._dir == 'b_top'

    @lazy_property
    def is_b_bottom(self):
        return self._dir == 'b_bottom'

    def validate(self):
        Element.validate(self)
        platform = self.parent_platform

        num_connections = len(list(self.connections(enabled=True)))

        if self.dtype not in Constants.TYPE_TO_SIZEOF.keys():
            self.add_error_message('Type "{}" is not a possible type.'.format(self.dtype))

        if self.is_sink and num_connections > 1:
            self.add_error_message('sink can have only one connection.')

    def rewrite(self):
        del self.dtype

        Element.rewrite(self)

    def connections(self, enabled=None):
        """Iterator over all connections to/from this port

        enabled: None for all, True for enabled only, False for disabled only
        """
        for con in self.parent_flowgraph.connections:
            #TODO clean this up - but how to get past this validation
            # things don't compare simply with an x in y because
            # bus ports are created differently.  
            port_in_con = False

            if self in con and (enabled is None or enabled == con.enabled):
                yield con




"""
Copyright (C) 2022 - Mahesh Patil <mbpatil@ee.iitb.ac.in>
This file is part of GSEIM.

GSEIM is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

class OutputBlock():
    def __init__(self, d_outparms={}, s_index_slv='', s_name='', d_output_block={}):
#       index_slv (string): index of the solve block to which this out block belongs

        self.d_parms = {}
        self.l_outvars = []

        if len(d_outparms) != 0:
            for k, v in d_outparms.items():
                self.d_parms[k] = v['default']

        if s_index_slv:
            self.index_slv = s_index_slv
        if s_name:
            self.name = s_name

        if len(d_output_block) != 0:
            self.name = d_output_block['name']
            self.index_slv = d_output_block['index_slv']

            for ov in d_output_block['outvars']:
                self.l_outvars.append(ov)

            for k in d_output_block['d_parms'].keys():
                self.d_parms[k] = d_output_block['d_parms'][k]

    def block_to_dict(self):
        d = {}
        d['name'] = self.name
        d['index_slv'] = self.index_slv
        d['outvars'] = self.l_outvars
        d['d_parms'] = self.d_parms
        return d



# Copyright 2016 Free Software Foundation, Inc.
# This file is part of GNU Radio
#
# GNU Radio Companion is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# GNU Radio Companion is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

from __future__ import absolute_import

import itertools
import re

from ..Constants import ADVANCED_PARAM_TAB
from ..utils import to_list
from ..Messages import send_warning

from .block import Block
from ._flags import Flags

def build(id, label='', category='', flags='', documentation='',
    parameters=None, inputs=None, outputs=None,
    e_left_nodes=None, e_right_nodes=None, e_top_nodes=None, e_bottom_nodes=None,
    b_left_nodes=None, b_right_nodes=None, b_top_nodes=None, b_bottom_nodes=None,
    **kwargs):

#   print('core/block/_build.py: build: id:',  id)

    block_id = id

    cls = type(str(block_id), (Block,), {})
    cls.key = block_id

    cls.label = label or block_id.title()
    cls.category = [cat.strip() for cat in category.split('/') if cat.strip()]

    cls.flags = Flags(flags)

    cls.documentation = {'': documentation.strip('\n\t ').replace('\\\n', '')}

    cls.inputs_data = build_ports(inputs, 'sink') if inputs else []
    cls.outputs_data = build_ports(outputs, 'source') if outputs else []

    cls.e_left_data = build_ports(e_left_nodes, 'e_left') if e_left_nodes else []
    cls.e_right_data = build_ports(e_right_nodes, 'e_right') if e_right_nodes else []
    cls.e_top_data = build_ports(e_top_nodes, 'e_top') if e_top_nodes else []
    cls.e_bottom_data = build_ports(e_bottom_nodes, 'e_bottom') if e_bottom_nodes else []

    cls.b_left_data = build_ports(b_left_nodes, 'b_left') if b_left_nodes else []
    cls.b_right_data = build_ports(b_right_nodes, 'b_right') if b_right_nodes else []
    cls.b_top_data = build_ports(b_top_nodes, 'b_top') if b_top_nodes else []
    cls.b_bottom_data = build_ports(b_bottom_nodes, 'b_bottom') if b_bottom_nodes else []

    cls.parameters_data = build_params(parameters or [],
         bool(cls.inputs_data), bool(cls.outputs_data), cls.flags, block_id)
    cls.extra_data = kwargs

    return cls

def build_ports(ports_raw, direction):
#   Note: direction is 'source' or 'sink'

    ports = []
    port_ids = set()
    stream_port_ids = itertools.count()

    for i, port_params in enumerate(ports_raw):
        port = port_params.copy()
        port['direction'] = direction

        port_id = port.setdefault('id', str(next(stream_port_ids)))
        if port_id in port_ids:
            raise Exception('Port id "{}" already exists in {}s'.format(port_id, direction))
        port_ids.add(port_id)

        ports.append(port)
    return ports

def build_params(params_raw, have_inputs, have_outputs, flags, block_id):
    params = []

    def add_param(**data):
        params.append(data)

    if flags.SHOW_ID in flags:
        add_param(id='id', name='ID', dtype='id', hide='none')
    else:
        add_param(id='id', name='ID', dtype='id', hide='all')

    base_params_n = {}
    for param_data in params_raw:
        param_id = param_data['id']
        if param_id in params:
            raise Exception('Param id "{}" is not unique'.format(param_id))

        base_key = param_data.get('base_key', None)
        param_data_ext = base_params_n.get(base_key, {}).copy()
        param_data_ext.update(param_data)

        add_param(**param_data_ext)
        base_params_n[param_id] = param_data_ext

    add_param(id='comment', name='Comment', dtype='_multiline', hide='part',
              default='', category=ADVANCED_PARAM_TAB)
    return params




# Copyright 2007, 2015 Free Software Foundation, Inc.
# This file is part of GNU Radio
#
# GNU Radio Companion is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# GNU Radio Companion is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

from __future__ import absolute_import

import traceback
import sys

#  A list of functions that can receive a message.
MESSENGERS_LIST = list()
_indent = ''

# Global FlowGraph Error and the file that caused it
flowgraph_error = None
flowgraph_error_file = None

def register_messenger(messenger):
    """
    Append the given messenger to the list of messengers.

    Args:
        messenger: a method that takes a string
    """
    MESSENGERS_LIST.append(messenger)

def set_indent(level=0):
    global _indent
    _indent = '    ' * level

def send(message):
    """
    Give the message to each of the messengers.

    Args:
        message: a message string
    """
    for messenger in MESSENGERS_LIST:
        messenger(_indent + message)

# register stdout by default
register_messenger(sys.stdout.write)

# Special functions for specific program functionalities
def send_init(platform):
    msg = "<<< Welcome to GSEIM (new) >>>\n\n" \
          "Block paths:\n\t{paths}\n"
    send(msg.format(
        config=platform.config,
        paths="\n\t".join(platform.config.block_paths))
    )

def send_start_load(file_path):
    send('\n ***Loading: "%s"\n' % file_path)

def send_error_msg_load(error):
    send('>>> Error: %s\n' % error)

def send_error_load(error):
    send_error_msg_load(error)
    traceback.print_exc()

def send_end_load():
    send('>>> Done\n')

def send_fail_load(error):
    send('Error: %s\n>>> Failure\n' % error)
    traceback.print_exc()

def send_start_gen(file_path):
    send('\nGenerating: %r\n' % file_path)

def send_auto_gen(file_path):
    send('>>> Messages.py: Generating: %r\n' % file_path)

def send_fail_gen(error):
    send('Messages.py: Generate Error: %s\n>>> Failure\n' % error)
    traceback.print_exc()

def send_start_exec(file_path):
    send('\nExecuting: %s\n' % file_path)

def send_verbose_exec(verbose):
    send(verbose)

def send_end_exec(code=0):
    send('\n>>> Done%s\n' % (" (return code %s)" % code if code else ""))

def send_fail_save(file_path):
    send('>>> Error: Cannot save: %s\n' % file_path)

def send_fail_connection(msg=''):
    send('>>> Error: Cannot create connection.\n' + ('\t{}\n'.format(msg) if msg else ''))

def send_fail_load_preferences(prefs_file_path):
    send('>>> Error: Cannot load preferences file: "%s"\n' % prefs_file_path)

def send_fail_save_preferences(prefs_file_path):
    send('>>> Error: Cannot save preferences file: "%s"\n' % prefs_file_path)

def send_warning(warning):
    send('>>> Warning: %s\n' % warning)



"""
Copyright (C) 2022 - Mahesh Patil <mbpatil@ee.iitb.ac.in>
This file is part of GSEIM.

GSEIM is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk

class AddOutputBlock(Gtk.Dialog):
    def __init__(self, parent, l_solve_blocks):
        Gtk.Dialog.__init__(self, title='Add Output Block', transient_for=parent, flags=0)
        self.add_buttons(
            Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL, Gtk.STOCK_OK, Gtk.ResponseType.OK
        )
        self.set_default_size(250, 100)
        self.set_position(Gtk.WindowPosition.CENTER)

        box = self.get_content_area()

        box.set_spacing(30)
        box.set_margin_top(5)

        grid = Gtk.Grid()

        grid.set_row_spacing(10)
        grid.set_column_spacing(10)

        label_1 = Gtk.Label(label = '  Choose parent solve block  ', xalign=0.5)

        self.combo_1 = Gtk.ComboBoxText()
        self.combo_1.set_entry_text_column(0)
        for slv in l_solve_blocks:
            self.combo_1.append_text(slv.name + ' (' + slv.index + ')')
        self.combo_1.set_active(0)

        grid.add(label_1)
        grid.attach_next_to(self.combo_1, label_1, Gtk.PositionType.BOTTOM, 1, 1)

        box.add(grid)
        self.show_all()

class PickOutputBlock(Gtk.Dialog):
    def __init__(self, parent, l_output_blocks, l_solve_blocks):
        Gtk.Dialog.__init__(self, title='Pick Output Block', transient_for=parent, flags=0)
        self.add_buttons(
            Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL, Gtk.STOCK_OK, Gtk.ResponseType.OK
        )
        self.set_default_size(250, 200)
        self.set_position(Gtk.WindowPosition.CENTER)

        grid = Gtk.Grid()
        grid.set_row_spacing(10)
        grid.set_column_spacing(10)
        grid.props.halign = Gtk.Align.CENTER

        scrolled_window = Gtk.ScrolledWindow()
        scrolled_window.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        scrolled_window.show()
        self.vbox.pack_start(scrolled_window, True, True, 0)
        scrolled_window.add(grid)

        self.l_names = []
        self.l_buttons = []

        for slv in l_solve_blocks:
            for out in slv.l_out:
                self.l_names.append(slv.name + ': ' + out)

        print('PickOutputBlock: self.l_names:', self.l_names)

        self.l_buttons.append(Gtk.RadioButton(label=self.l_names[0]))

        for i in range(1, len(self.l_names)):
            button1 = Gtk.RadioButton.new_from_widget(self.l_buttons[i-1])
            button1.set_label(self.l_names[i])
            self.l_buttons.append(button1)

        grid.add(self.l_buttons[0])

        for i in range(1, len(self.l_names)):
            grid.attach_next_to(self.l_buttons[i], self.l_buttons[i-1],
               Gtk.PositionType.BOTTOM, 1, 1)

        self.show_all()

class EditOutputBlock(Gtk.Dialog):
#   Present the selected output block for editing:

    def __init__(self, parent, blk, d_lib, l_outvars, l_solve_blocks):
#       blk is the selected output block

        i_slv = int(blk.index_slv)
        s1 = 'Block ' + blk.name + ' in Solve Block ' \
           + l_solve_blocks[i_slv].name + '(' \
           + l_solve_blocks[i_slv].index + ')'

        Gtk.Dialog.__init__(self,
            title=s1,
            transient_for=parent, flags=0)
        self.add_buttons(
            Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL, Gtk.STOCK_OK, Gtk.ResponseType.OK
        )
        self.set_default_size(400, 800)
        self.set_position(Gtk.WindowPosition.CENTER)

        grid = Gtk.Grid()
        grid.set_row_spacing(0)
        grid.set_column_spacing(0)

        scrolled_window = Gtk.ScrolledWindow()
        scrolled_window.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        scrolled_window.show()
        self.vbox.pack_start(scrolled_window, True, True, 0)

        scrolled_window.add(grid)

        d_widgets = {}
        set1 = {'yes', 'no'}

        self.d_widgets_1 = {}

        for k, v in d_lib.items():
            l_options = v['options']
            s_value = blk.d_parms[k]

            label_1 = Gtk.Label(label = '  ' + k + '  ', xalign=0)
            l_widgets = []
            l_widgets.append([label_1, [0, 0, 1, 1]])

            if l_options[0] == 'none':
                if s_value == 'none':
                    entry_1 = Gtk.Entry(text='')
                else:
                    entry_1 = Gtk.Entry(text=s_value)

                self.d_widgets_1[k] = {'type_widget': 'entry', 'widget':entry_1}
                l_widgets.append([entry_1, [1, 0, 1, 1]])
            else:
                if set(l_options) == set1:
                    button_1 = Gtk.CheckButton()
                    button_1.set_active(s_value == 'yes')
                    button_1.set_label('yes/no')
                    l_widgets.append([button_1, [1, 0, 1, 1]])
                    self.d_widgets_1[k] = {'type_widget': 'checkbutton', 'widget':button_1}
                else:
                    combo_1 = Gtk.ComboBoxText()
                    combo_1.set_entry_text_column(0)
                    for x in l_options:
                        combo_1.append_text(x)
                    combo_1.set_active(l_options.index(s_value))
                    self.d_widgets_1[k] = {'type_widget': 'combo', 'widget':combo_1}

                    l_widgets.append([combo_1, [1, 0, 1, 1]])

            d_widgets[k] = l_widgets

        l_keys = []
        for k in d_lib.keys():
            l_keys.append(k)

        disp = 0

        for k in l_keys:
            for l0 in d_widgets[k]:
                l1 = l0[1]
                grid.attach(l0[0], l1[0], l1[1] + disp, l1[2], l1[3])
            d0 = max(map(lambda x: x[1][1], d_widgets[k])) + 1
                    
            disp += d0

        n_ov = len(l_outvars)
        label_ov = Gtk.Label(label = "  Output Variables  ", xalign=0.5)
        grid.attach(label_ov, 0, disp, 1, n_ov)

        self.l_ov_buttons = []

        for i, ov_name in enumerate(l_outvars):
            button_1 = Gtk.CheckButton()
            button_1.set_active(ov_name in blk.l_outvars)
            button_1.set_label(ov_name)
            self.l_ov_buttons.append(button_1)
            grid.attach(button_1, 1, disp + i, 1, 1)

        disp += n_ov

        self.show_all()



"""
Copyright 2007, 2008, 2009 Free Software Foundation, Inc.
This file is part of GNU Radio

GNU Radio Companion is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

GNU Radio Companion is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""

from __future__ import absolute_import, division

from argparse import Namespace
from math import pi

from . import colors
from .drawable import Drawable
from .. import Utils
from ..Constants import (
    CONNECTOR_ARROW_BASE,
    CONNECTOR_ARROW_HEIGHT,
    GR_MESSAGE_DOMAIN,
    LINE_SELECT_SENSITIVITY,
    CONNECTOR_LINE_WIDTH,
    CONNECTOR_LINE_WIDTH_FACTOR,
)
from ...core.Connection import Connection as CoreConnection
from ...core.utils.descriptors import nop_write

import sys

import cairo
from .. import Constants

from . import rect_wiring as rwire

class Connection(CoreConnection, Drawable):
    """
    A graphical connection for ports.
    The connection has 2 parts, the arrow and the wire.
    The coloring of the arrow and wire exposes the status of 3 states:
        enabled/disabled, valid/invalid, highlighted/non-highlighted.
    The wire coloring exposes the enabled and highlighted states.
    The arrow coloring exposes the enabled and valid states.
    """

    def __init__(self, *args, **kwargs):
        super(self.__class__, self).__init__(*args, **kwargs)
        Drawable.__init__(self)

        self._line = []
        self._line_width_factor = 1.0
        self._color1 = self._color2 = None

        self._current_port_rotations = self._current_coordinates = None

        self._rel_points = None  # connection coordinates relative to sink/source
        self._arrow_rotation = 0.0  # rotation of the arrow in radians
        self._current_cr = None  # for what_is_selected() of curved line
        self._line_path = None

        self._l_wire = []
        self._l_wire_src = []
        self._l_arrow = []
        self._l_arrow_src = []

    @nop_write
    @property
    def coordinate(self):
        return self.source_port.connector_coordinate_absolute

    @nop_write
    @property
    def rotation(self):
        """
        Get the 0 degree rotation.
        Rotations are irrelevant in connection.

        Returns:
            0
        """
        return 0

    def create_shapes(self):
        """Pre-calculate relative coordinates."""

        source = self.source_port
        sink = self.sink_port
        rotate = Utils.get_rotated_coordinate

        # first two components relative to source connector, rest relative to sink connector

        self.wiring_style = source.parent_flowgraph.options_block.params['wiring_style'].get_value()

        source_mirror = source.parent_block.mirror
        source_rotation = source.parent_block.rotation

        if 'Namespace' not in str(type(sink.parent_block)):
            sink_rotation = sink.parent_block.rotation
            sink_mirror = sink.parent_block.mirror
        else:
            sink_rotation = 0
            sink_mirror = 'none'

        if self.wiring_style == 'curved':
            if source.port_type == 'flowgraph':
                if source.parent_block.name.startswith('connector_f'):
                    index1 = source.port_index
                    pos = source.parent_block.params['output' + str(int(index1)+1)].get_value()

                    if pos == 'top':
                        px_source = ((  0,-15), (  0,-50))
                    elif pos == 'bottom':
                        px_source = ((  0, 15), (  0, 50))
                    elif pos == 'right':
                        px_source = (( 15,  0), ( 50,  0))

                    self._rel_points = [
                        rotate(px_source[0], source.rotation),
                        rotate(px_source[1], source.rotation),
                        rotate((-50, 0), sink.rotation),
                        rotate((-15, 0), sink.rotation),
                        rotate((-CONNECTOR_ARROW_HEIGHT, 0), sink.rotation),
                    ]
                    r_sink = sink.rotation
                else:
                    if source_mirror == 'none' and sink_mirror == 'none':
                        self._rel_points = [
                            rotate((15, 0), source.rotation),  # line from 0,0 to here, bezier curve start
                            rotate((50, 0), source.rotation),  # bezier curve control point 1
                            rotate((-50, 0), sink.rotation),  # bezier curve control point 2
                            rotate((-15, 0), sink.rotation),  # bezier curve end
                            rotate((-CONNECTOR_ARROW_HEIGHT, 0), sink.rotation),  # line to arrow head
                        ]
                        r_sink = sink.rotation
                    else:
                        r_source = (180 + source.rotation) % 360 if source_mirror == 'v' else source.rotation
                        r_sink = (180 + sink.rotation) % 360 if sink_mirror == 'v' else sink.rotation
                        self._rel_points = [
                            rotate((15, 0), r_source),
                            rotate((50, 0), r_source),
                            rotate((-50, 0), r_sink),
                            rotate((-15, 0), r_sink),
                            rotate((-CONNECTOR_ARROW_HEIGHT, 0), r_sink),
                        ]
            elif source.port_type == 'electrical':
                if source_mirror == 'none':
                    px_source = {
                       'e_right' :(( 15,  0), ( 50,  0)),
                       'e_left'  :((-15,  0), (-50,  0)),
                       'e_top'   :((  0,-15), (  0,-50)),
                       'e_bottom':((  0, 15), (  0, 50)),
                    }[source.port_subtype]
                elif source_mirror == 'v':
                    px_source = {
                       'e_left'  :(( 15,  0), ( 50,  0)),
                       'e_right' :((-15,  0), (-50,  0)),
                       'e_top'   :((  0,-15), (  0,-50)),
                       'e_bottom':((  0, 15), (  0, 50)),
                    }[source.port_subtype]
                elif source_mirror == 'h':
                    px_source = {
                       'e_right' :(( 15,  0), ( 50,  0)),
                       'e_left'  :((-15,  0), (-50,  0)),
                       'e_bottom':((  0,-15), (  0,-50)),
                       'e_top'   :((  0, 15), (  0, 50)),
                    }[source.port_subtype]

                if sink_mirror == 'none':
                    px_sink = {
                       'e_right' :(( 50,  0), ( 15,  0)),
                       'e_left'  :((-50,  0), (-15,  0)),
                       'e_top'   :((  0,-50), (  0,-15)),
                       'e_bottom':((  0, 50), (  0, 15)),
                    }[sink.port_subtype]
                elif sink_mirror == 'v':
                    px_sink = {
                       'e_left'  :(( 50,  0), ( 15,  0)),
                       'e_right' :((-50,  0), (-15,  0)),
                       'e_top'   :((  0,-50), (  0,-15)),
                       'e_bottom':((  0, 50), (  0, 15)),
                    }[sink.port_subtype]
                elif sink_mirror == 'h':
                    px_sink = {
                       'e_right' :(( 50,  0), ( 15,  0)),
                       'e_left'  :((-50,  0), (-15,  0)),
                       'e_bottom':((  0,-50), (  0,-15)),
                       'e_top'   :((  0, 50), (  0, 15)),
                    }[sink.port_subtype]

                if source_mirror == 'none' and sink_mirror == 'none':
                    self._rel_points = [
                        rotate(px_source[0], source.rotation),
                        rotate(px_source[1], source.rotation),
                        rotate(px_sink[0], sink.rotation),
                        rotate(px_sink[1], sink.rotation),
                        rotate((0, 0), sink.rotation),
                    ]
                else:
                    r_source = (180 + source.rotation) % 360 if source_mirror == 'h' else source.rotation
                    r_sink = (180 + sink.rotation) % 360 if sink_mirror == 'h' else sink.rotation
                    self._rel_points = [
                        rotate(px_source[0], r_source),
                        rotate(px_source[1], r_source),
                        rotate(px_sink[0], r_sink),
                        rotate(px_sink[1], r_sink),
                        rotate((0, 0), r_sink),
                    ]
            elif source.port_type == 'bus':
                if source_mirror == 'none':
                    px_source = {
                       'b_right' :(( 15,  0), ( 50,  0)),
                       'b_left'  :((-15,  0), (-50,  0)),
                       'b_top'   :((  0,-15), (  0,-50)),
                       'b_bottom':((  0, 15), (  0, 50)),
                    }[source.port_subtype]
                elif source_mirror == 'v':
                    px_source = {
                       'b_left'  :(( 15,  0), ( 50,  0)),
                       'b_right' :((-15,  0), (-50,  0)),
                       'b_top'   :((  0,-15), (  0,-50)),
                       'b_bottom':((  0, 15), (  0, 50)),
                    }[source.port_subtype]
                elif source_mirror == 'h':
                    px_source = {
                       'b_right' :(( 15,  0), ( 50,  0)),
                       'b_left'  :((-15,  0), (-50,  0)),
                       'b_bottom':((  0,-15), (  0,-50)),
                       'b_top'   :((  0, 15), (  0, 50)),
                    }[source.port_subtype]

                if sink_mirror == 'none':
                    px_sink = {
                       'b_right' :(( 50,  0), ( 15,  0)),
                       'b_left'  :((-50,  0), (-15,  0)),
                       'b_top'   :((  0,-50), (  0,-15)),
                       'b_bottom':((  0, 50), (  0, 15)),
                    }[sink.port_subtype]
                elif sink_mirror == 'v':
                    px_sink = {
                       'b_left'  :(( 50,  0), ( 15,  0)),
                       'b_right' :((-50,  0), (-15,  0)),
                       'b_top'   :((  0,-50), (  0,-15)),
                       'b_bottom':((  0, 50), (  0, 15)),
                    }[sink.port_subtype]
                elif sink_mirror == 'h':
                    px_sink = {
                       'b_right' :(( 50,  0), ( 15,  0)),
                       'b_left'  :((-50,  0), (-15,  0)),
                       'b_bottom':((  0,-50), (  0,-15)),
                       'b_top'   :((  0, 50), (  0, 15)),
                    }[sink.port_subtype]

                if source_mirror == 'none' and sink_mirror == 'none':
                    self._rel_points = [
                        rotate(px_source[0], source.rotation),
                        rotate(px_source[1], source.rotation),
                        rotate(px_sink[0], sink.rotation),
                        rotate(px_sink[1], sink.rotation),
                        rotate((0, 0), sink.rotation),
                    ]
                else:
                    r_source = (180 + source.rotation) % 360 if source_mirror == 'h' else source.rotation
                    r_sink = (180 + sink.rotation) % 360 if sink_mirror == 'h' else sink.rotation
                    self._rel_points = [
                        rotate(px_source[0], r_source),
                        rotate(px_source[1], r_source),
                        rotate(px_sink[0], r_sink),
                        rotate(px_sink[1], r_sink),
                        rotate((0, 0), r_sink),
                    ]

            if source.port_type == 'flowgraph':
                self._arrow_rotation = -r_sink / 180 * pi

        self._current_coordinates = None  # triggers _make_path()

        self._line_width_factor = 1.0
        self._color1 = colors.CONNECTION_ENABLED_COLOR
        self._color2 = colors.CONNECTION_ENABLED_COLOR

        if not self._bounding_points:
            if self.wiring_style == 'curved':
                self._make_path()  # no cr set --> only sets bounding_points for extent
            elif self.wiring_style == 'rectilinear':
#               self._bounding_points = tuple(self._l_wire_src)
                if self._l_wire_src:
                    self._bounding_points = tuple(self._l_wire_src)
                else:
                    p0 = 0, 0
                    self._bounding_points = p0, p0, p0, p0

    def _make_path(self, cr=None):
        if self.wiring_style == 'curved':
            source = self.source_port
            sink = self.sink_port

            x_pos, y_pos = self.source_port.connector_coordinate_absolute
            # x_start, y_start = self.source_port.get_connector_coordinate()
            x_end, y_end = self.sink_port.connector_coordinate_absolute

            # sink connector relative to sink connector
            x_e, y_e = x_end - x_pos, y_end - y_pos

            # make rel_point all relative to source connector
            p0 = 0, 0  # x_start - x_pos, y_start - y_pos
            p1, p2, (dx_e1, dy_e1), (dx_e2, dy_e2), (dx_e3, dy_e3) = self._rel_points
            p3 = x_e + dx_e1, y_e + dy_e1
            p4 = x_e + dx_e2, y_e + dy_e2
            p5 = x_e + dx_e3, y_e + dy_e3
            self._bounding_points = p0, p1, p4, p5  # ignores curved part =(

            if cr:
                cr.move_to(*p0)
                cr.line_to(*p1)
                cr.curve_to(*(p2 + p3 + p4))

                cr.line_to(*p5)
                self._line_path = cr.copy_path()

        elif self.wiring_style == 'rectilinear':
            source = self.source_port
            sink = self.sink_port

            if 'Namespace' in str(type(sink.parent_block)):
                if not self._bounding_points:
                    p0 = 0, 0
                    self._bounding_points = p0, p0, p0, p0
                return
#           end debug

            _delta1 = Constants.RECT_WIRING_DELTA1
            _wc = source.width

            _b1_coord = source.parent_block.coordinate

            if source.rotation in (0, 180):
                _b1_L = source.parent_block.width
                _b1_W = source.parent_block.height
            else:
                _b1_L = source.parent_block.height
                _b1_W = source.parent_block.width

            _b2_coord = sink.parent_block.coordinate

            if sink.rotation in (0, 180):
                _b2_L = sink.parent_block.width
                _b2_W = sink.parent_block.height
            else:
                _b2_L = sink.parent_block.height
                _b2_W = sink.parent_block.width

            _b1 = rwire.make_block_1(_b1_coord, _b1_L, _b1_W, _wc, _delta1)
            _b2 = rwire.make_block_1(_b2_coord, _b2_L, _b2_W, _wc, _delta1)

            d1 = {
              'sink'    : 0,
              'e_left'  : 0,
              'b_left'  : 0,
              'bottom'  : 1,
              'e_bottom': 1,
              'b_bottom': 1,
              'source'  : 2,
              'right'   : 2,
              'e_right' : 2,
              'b_right' : 2,
              'top'     : 3,
              'e_top'   : 3,
              'b_top'   : 3,
            }

            source_mirror = source.parent_block.mirror
            source_rotation = source.parent_block.rotation

            if 'Namespace' not in str(type(sink.parent_block)):
                sink_rotation = sink.parent_block.rotation
                sink_mirror = sink.parent_block.mirror
            else:
                sink_rotation = 0
                sink_mirror = 'none'

            if source.parent_block.name.startswith('connector_f'):
                index1 = source.port_index
                source_subtype = source.parent_block.params['output' + str(int(index1)+1)].get_value()
            else:
                source_subtype = 'source' if source.port_subtype == '' else source.port_subtype
                if source_mirror != 'none':
                    source_subtype = {
                       ('source'  , 'v'): 'sink',
                       ('sink'    , 'v'): 'source',
                       ('source'  , 'h'): 'source',
                       ('sink'    , 'h'): 'sink',
                       ('e_left'  , 'v'): 'e_right',
                       ('e_right' , 'v'): 'e_left',
                       ('e_left'  , 'h'): 'e_left',
                       ('e_right' , 'h'): 'e_right',
                       ('e_top'   , 'v'): 'e_top',
                       ('e_bottom', 'v'): 'e_bottom',
                       ('e_top'   , 'h'): 'e_bottom',
                       ('e_bottom', 'h'): 'e_top',
                       ('b_left'  , 'v'): 'b_right',
                       ('b_right' , 'v'): 'b_left',
                       ('b_left'  , 'h'): 'b_left',
                       ('b_right' , 'h'): 'b_right',
                       ('b_top'   , 'v'): 'b_top',
                       ('b_bottom', 'v'): 'b_bottom',
                       ('b_top'   , 'h'): 'b_bottom',
                       ('b_bottom', 'h'): 'b_top',
                    }[(source_subtype, source_mirror)]

            sink_subtype = 'sink' if sink.port_subtype == '' else sink.port_subtype
            if sink_mirror != 'none':
                sink_subtype = {
                   ('source'  , 'v'): 'sink',
                   ('sink'    , 'v'): 'source',
                   ('source'  , 'h'): 'source',
                   ('sink'    , 'h'): 'sink',
                   ('e_left'  , 'v'): 'e_right',
                   ('e_right' , 'v'): 'e_left',
                   ('e_left'  , 'h'): 'e_left',
                   ('e_right' , 'h'): 'e_right',
                   ('e_top'   , 'v'): 'e_top',
                   ('e_bottom', 'v'): 'e_bottom',
                   ('e_top'   , 'h'): 'e_bottom',
                   ('e_bottom', 'h'): 'e_top',
                   ('b_left'  , 'v'): 'b_right',
                   ('b_right' , 'v'): 'b_left',
                   ('b_left'  , 'h'): 'b_left',
                   ('b_right' , 'h'): 'b_right',
                   ('b_top'   , 'v'): 'b_top',
                   ('b_bottom', 'v'): 'b_bottom',
                   ('b_top'   , 'h'): 'b_bottom',
                   ('b_bottom', 'h'): 'b_top',
                }[(sink_subtype, sink_mirror)]

            kp_source, kp_sink = [d1[x] for x in (source_subtype, sink_subtype)]

            d2 = { 0:0, 90:1, 180:2, 270:3}
            l_port_dir = ['W', 'S', 'E', 'N']

            k_source = (d2[source.rotation] + kp_source) % 4

            k_sink = (d2[sink.rotation] + kp_sink) % 4

            _dir1 = l_port_dir[k_source]
            _dir2 = l_port_dir[k_sink  ]

            _cc1 = source.connector_coordinate_absolute
            _cc2 = sink.connector_coordinate_absolute

            _arrow_h = 8 
            _arrow_w = 3 

            _draw_arrow1 = False
            _draw_arrow2 = sink.port_type == 'flowgraph'

            self._l_wire, self._l_arrow = rwire.get_wire(
                _b1, _cc1, _dir1, _draw_arrow1,
                _b2, _cc2, _dir2, _draw_arrow2,
                _delta1, _arrow_h, _arrow_w)

            x0, y0 = self._l_wire[0]

            self._l_wire_src = []
            for x, y in self._l_wire:
                self._l_wire_src.append((x-x0, y-y0))

            if sink.port_type == 'flowgraph':
                self._l_arrow_src = []
                for x, y in self._l_arrow:
                    self._l_arrow_src.append((x-x0, y-y0))

#           debug
#           if not self._bounding_points:
#               p0 = 0, 0
#               self._bounding_points = p0, p0, p0, p0
#           end debug

            if cr:
                cr.move_to(*self._l_wire_src[0])
                for k in range(1,len(self._l_wire_src)):
                    cr.line_to(*self._l_wire_src[k])

                self._line_path = cr.copy_path()

    def draw(self, cr):
        """
        Draw the connection.
        """
        self._current_cr = cr
        sink = self.sink_port
        source = self.source_port

        port_rotations = (source.rotation, sink.rotation)
        if self._current_port_rotations != port_rotations:
            self.create_shapes()  # triggers _make_path() call below
            self._current_port_rotations = port_rotations

        new_coordinates = (source.parent_block.coordinate, sink.parent_block.coordinate)
        if self._current_coordinates != new_coordinates:
            self._make_path(cr)
            self._current_coordinates = new_coordinates

        color1, color2 = (
            None if color is None else
            colors.HIGHLIGHT_COLOR if self.highlighted else
            colors.CONNECTION_DISABLED_COLOR if not self.enabled else
            colors.CONNECTION_ERROR_COLOR if not self.is_valid() else
            color
            for color in (self._color1, self._color2)
        )

        if self.wiring_style == 'curved':
            cr.translate(*self.coordinate)
            cr.set_line_width(self._line_width_factor * cr.get_line_width())
            cr.new_path()
            cr.append_path(self._line_path)

            if source.port_type == 'flowgraph':
                arrow_pos = cr.get_current_point()

            if color1:  # not a message connection
                cr.set_source_rgba(*color1)
                cr.stroke_preserve()

            if color1 != color2:
                cr.save()
                cr.set_dash([5.0, 5.0], 5.0 if color1 else 0.0)
                cr.set_source_rgba(*color2)
                cr.stroke()
                cr.restore()
            else:
                cr.new_path()

            if source.port_type == 'flowgraph':
                cr.move_to(*arrow_pos)
                cr.set_source_rgba(*color2)
                cr.rotate(self._arrow_rotation)
                cr.rel_move_to(CONNECTOR_ARROW_HEIGHT, 0)
                cr.rel_line_to(-CONNECTOR_ARROW_HEIGHT, -CONNECTOR_ARROW_BASE/2)
                cr.rel_line_to(0, CONNECTOR_ARROW_BASE)
                cr.close_path()
                cr.fill()

        elif self.wiring_style == 'rectilinear':
            cr.set_line_join(cairo.LineJoin.ROUND)

            cr.translate(*self.coordinate)

#           this is required to ensure that canvas/port.py uses the same line width
#           in drawing connector ports. Not the best perhaps, but will do...

            Constants.temp1 = CONNECTOR_LINE_WIDTH * cr.get_line_width()
            cr.set_line_width(Constants.temp1)
            cr.new_path()

#           debug
            if 'Namespace' not in str(type(sink.parent_block)):
                cr.append_path(self._line_path)
#           cr.append_path(self._line_path)
#           end debug

            if color1:  # not a message connection
                cr.set_source_rgba(*color1)
                cr.stroke_preserve()

            if color1 != color2:
                cr.save()
                cr.set_dash([5.0, 5.0], 5.0 if color1 else 0.0)
                cr.set_source_rgba(*color2)
                cr.stroke()
                cr.restore()
            else:
                cr.new_path()

            if source.port_type == 'flowgraph':

                if 'Namespace' not in str(type(sink.parent_block)):
                    if not sink.parent_block.name.split('$')[0].startswith('connector_f'):
                        cr.move_to(*self._l_arrow_src[0])
                        cr.set_source_rgba(*color2)
                        cr.line_to(*self._l_arrow_src[1])
                        cr.line_to(*self._l_arrow_src[2])
                        cr.close_path()
                        cr.fill()

    def what_is_selected(self, coor, coor_m=None):
        """
        Returns:
            self if one of the areas/lines encompasses coor, else None.
        """

        if coor_m:
            return Drawable.what_is_selected(self, coor, coor_m)

        x, y = [a - b for a, b in zip(coor, self.coordinate)]

        cr = self._current_cr

        if cr is None:
            return
        cr.save()
        cr.new_path()
        cr.append_path(self._line_path)
        cr.set_line_width(cr.get_line_width() * LINE_SELECT_SENSITIVITY)
#       cr.set_line_width(cr.get_line_width())
        hit = cr.in_stroke(x, y)
        cr.restore()

        if hit:
            return self

class DummyCoreConnection(object):
    def __init__(self, source_port, **kwargs):
        self.parent_platform = source_port.parent_platform
        self.source_port = source_port

        self.sink_port = self._dummy_port = Namespace(
            domain=source_port.domain,
            rotation=0,
            coordinate=(0, 0),
            connector_coordinate_absolute=(0, 0),
            connector_direction=0,
            parent_block=Namespace(coordinate=(0, 0)),
            port_subtype='',
        )

        self.enabled = True
        self.highlighted = False,
        self.is_valid = lambda: True
        self.update(**kwargs)

    def update(self, coordinate=None, rotation=None, sink_port=None):
        dp = self._dummy_port
        self.sink_port = sink_port if sink_port else dp
        if coordinate:
            dp.coordinate = coordinate
            dp.connector_coordinate_absolute = coordinate
            dp.parent_block.coordinate = coordinate
        if rotation is not None:
            dp.rotation = rotation
            dp.connector_direction = (180 + rotation) % 360

    @property
    def has_real_sink(self):
        return self.sink_port is not self._dummy_port

DummyConnection = Connection.make_cls_with_base(DummyCoreConnection)



# Copyright 2016 Free Software Foundation, Inc.
# This file is part of GNU Radio
#
# GNU Radio Companion is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# GNU Radio Companion is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

from __future__ import absolute_import

class FlowGraphProxy(object):  # TODO: move this in a refactored Generator

    def __init__(self, fg):
        self.orignal_flowgraph = fg

    def __getattr__(self, item):
        return getattr(self.orignal_flowgraph, item)

    def get_pad_sources(self):
        """
        Get a list of pad source blocks sorted by id order.

        Returns:
            a list of pad source blocks in this flow graph
        """
        pads = [b for b in self.get_enabled_blocks() if b.key == 'pad_source']
        return sorted(pads, key=lambda x: int(x.name.split('$')[-1]))

    def get_pad_sinks(self):
        """
        Get a list of pad sink blocks sorted by id order.

        Returns:
            a list of pad sink blocks in this flow graph
        """
        pads = [b for b in self.get_enabled_blocks() if b.key == 'pad_sink']
        return sorted(pads, key=lambda x: int(x.name.split('$')[-1]))

    def get_pad_e_lefts(self):
        pads = [b for b in self.get_enabled_blocks() if b.key == 'pad_e_left']
        return sorted(pads, key=lambda x: int(x.name.split('$')[-1]))

    def get_pad_e_rights(self):
        pads = [b for b in self.get_enabled_blocks() if b.key == 'pad_e_right']
        return sorted(pads, key=lambda x: int(x.name.split('$')[-1]))

    def get_pad_e_tops(self):
        pads = [b for b in self.get_enabled_blocks() if b.key == 'pad_e_top']
        return sorted(pads, key=lambda x: int(x.name.split('$')[-1]))

    def get_pad_e_bottoms(self):
        pads = [b for b in self.get_enabled_blocks() if b.key == 'pad_e_bottom']
        return sorted(pads, key=lambda x: int(x.name.split('$')[-1]))

    def get_pad_b_lefts(self):
        pads = [b for b in self.get_enabled_blocks() if b.key == 'pad_b_left']
        return sorted(pads, key=lambda x: int(x.name.split('$')[-1]))

    def get_pad_b_rights(self):
        pads = [b for b in self.get_enabled_blocks() if b.key == 'pad_b_right']
        return sorted(pads, key=lambda x: int(x.name.split('$')[-1]))

    def get_pad_b_tops(self):
        pads = [b for b in self.get_enabled_blocks() if b.key == 'pad_b_top']
        return sorted(pads, key=lambda x: int(x.name.split('$')[-1]))

    def get_pad_b_bottoms(self):
        pads = [b for b in self.get_enabled_blocks() if b.key == 'pad_b_bottom']
        return sorted(pads, key=lambda x: int(x.name.split('$')[-1]))




"""
Copyright 2007, 2008, 2009 Free Software Foundation, Inc.
This file is part of GNU Radio

GNU Radio Companion is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

GNU Radio Companion is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""

from __future__ import absolute_import, division

import math
import sys

from gi.repository import Gtk, Pango, PangoCairo

from . import colors
from .drawable import Drawable
from .. import Actions, Utils, Constants
from ..Constants import (
    BLOCK_LABEL_PADDING, PORT_SPACING, PORT_SEPARATION, LABEL_SEPARATION,
    PORT_BORDER_SEPARATION, BLOCK_FONT, PARAM_FONT, SHOW_TEXT_LABEL_SEPARATION,
    TEXT_BLOCK_LABEL_PADDING, BLOCK_LABEL_PADDING_TAGS, RECT_WIRING_DELTA1
)
from ...core import utils
from ...core.blocks import Block as CoreBlock

from ...core.utils import gutils as gu
import cairo

class Block(CoreBlock, Drawable):
    """The graphical signal block."""

    def __init__(self, parent, **n):
        """
        Block constructor.
        Add graphics related params to the block.
        """
        super(self.__class__, self).__init__(parent, **n)

        self.states.update(coordinate=(0, 0), rotation=0)
        self.width = self.height = 0
        Drawable.__init__(self)  # needs the states and initial sizes

        self._surface_layouts = [
            None,  # title
            None,  # params
        ]
        self._surface_layouts_offsets = 0, 0
        self._comment_layout = None

        self._area = []
        self._border_color = self._bg_color = colors.BLOCK_ENABLED_COLOR
        self._font_color = list(colors.FONT_COLOR)

        self.drawing_scheme = self.params['drawing_scheme'].get_value()

        self.l_coord = []
        self.l_draw = []
        self.c_ = []
        self.a_ = []
        self.s_ = [None, None, None, None]
        self.t_ = []

        self.c__ = []
        self.a__ = []
        self.s__ = self.s_
        self.has_arc = False

        self.port_offset_l = 0
        self.port_offset_r = 0
        self.port_offset_t = 0
        self.port_offset_b = 0

        if self.category:
            if 'Hier' in self.category[0]:
                self.file_symbol = parent.parent.config.hier_block_lib_dir + '/' + self.key + '.symbol.py'
            else:
                self.file_symbol = parent.parent.config.block_lib_dir + '/' + self.key + '.symbol.py'
        else:
            self.file_symbol = parent.parent.config.block_lib_dir + '/' + self.key + '.symbol.py'

        self.port_sep_x = int(self.params['port_sep_x'].get_value())
        self.port_sep_y = int(self.params['port_sep_y'].get_value())
        self.port_block_x = int(self.params['port_block_x'].get_value())
        self.port_block_y = int(self.params['port_block_y'].get_value())

        if self.drawing_scheme == 'symbol':
#           if the block name is b_1, assume that the python code for the symbol
#           is in b_1.symbol.py (in the blocks/ directory)

            self.port_offset_l = int(self.params['port_offset_l'].get_value())
            self.port_offset_r = int(self.params['port_offset_r'].get_value())
            self.port_offset_t = int(self.params['port_offset_t'].get_value())
            self.port_offset_b = int(self.params['port_offset_b'].get_value())

            self.l_coord = gu.python_code_to_list(
                filename=self.file_symbol,
                keyword1='begin_coord',
                keyword2='end_coord')

            for s in self.l_coord:
                exec(s, None, globals())
            self.c_ = c_
            self.a_ = a_
            self.s_ = s_
            self.t_ = t_

            self.l_draw = gu.python_code_to_list(
                filename=self.file_symbol,
                keyword1='begin_draw',
                keyword2='end_draw')

            self.has_arc = any('arc' in x for x in self.l_draw)

        self._hovering = False
        self.label_layout = None

        self.parent0 = parent

    @property
    def coordinate(self):
        """
        Get the coordinate from the position param.

        Returns:
            the coordinate tuple (x, y) or (0, 0) if failure
        """
        return Utils.scale(self.states['coordinate'])

    @coordinate.setter
    def coordinate(self, coor):
        """
        Set the coordinate into the position param.

        Args:
            coor: the coordinate tuple (x, y)
        """
        coor = Utils.scale(coor, reverse=True)
        if Actions.TOGGLE_SNAP_TO_GRID.get_active():

            coor = Utils.align_to_grid(coor)

        self.states['coordinate'] = coor

    @property
    def rotation(self):
        """
        Get the rotation from the position param.

        Returns:
            the rotation in degrees or 0 if failure
        """
        return self.states['rotation']

    @rotation.setter
    def rotation(self, rot):
        """
        Set the rotation into the position param.

        Args:
            rot: the rotation in degrees
        """
        self.states['rotation'] = rot

    def _update_colors(self):

        block_type = type(self).__name__
        if block_type.startswith('dummy'):
            color1 = colors.BLOCK_DUMMY_COLOR
        elif block_type.startswith('pad'):
            color1 = colors.BLOCK_PAD_COLOR
        elif block_type.startswith('bus'):
            color1 = colors.BLOCK_BUS_COLOR
        else:
            color1 = colors.BLOCK_ENABLED_COLOR

        if self.name.startswith('connector_f'):
            self._bg_color = colors.F_PORT_COLOR
        elif self.name.startswith('connector_e'):
            self._bg_color = colors.E_PORT_COLOR
        elif self.name.startswith('connector_b'):
            self._bg_color = colors.B_PORT_COLOR
        else:
            self._bg_color = color1

        self._font_color[-1] = 1.0 if self.state == 'enabled' else 0.4

        self._border_color = (
            colors.MISSING_BLOCK_BORDER_COLOR if self.is_dummy_block else
            colors.BORDER_COLOR_DISABLED if not self.state == 'enabled' else colors.BORDER_COLOR
        )

    def get_offset(self, _rotate_strict, side, theta, port_offset, U, N, h1, k):

        if _rotate_strict == 'no':
            _flag1 = 'normal'
        elif theta == 0:
            _flag1 = 'normal'
        else:
            _flag1 = {
               (  'vertical',  90):'normal',
               (  'vertical', 180):'reverse',
               (  'vertical', 270):'reverse',
               ('horizontal',  90):'reverse',
               ('horizontal', 180):'reverse',
               ('horizontal', 270):'normal',
            }[(side, theta)]

        del1 = Constants.CANVAS_GRID_SIZE

        port_offs = del1*port_offset

        if self.drawing_scheme == 'symbol':
          if theta == 0:
            port_offset_1 = port_offs
          elif theta == 180:
            port_offset_1 = -port_offs
          elif theta == 90:
            if side == 'horizontal':
              port_offset_1 = -port_offs
            else:
              port_offset_1 = port_offs
          elif theta == 270:
            if side == 'horizontal':
              port_offset_1 = port_offs
            else:
              port_offset_1 = -port_offs

        if self.drawing_scheme == 'symbol':

            if side == 'vertical':
                del_port = 2*del1*self.port_sep_y
            elif side == 'horizontal':
                del_port = 2*del1*self.port_sep_x
            if _flag1 == 'normal':
                _offset = (U - (N-1)*del_port - h1)/2 + k*del_port + port_offset_1
            elif _flag1 == 'reverse':
                _offset = (U - (N-1)*del_port - h1)/2 + (N-k-1)*del_port + port_offset_1

        elif self.drawing_scheme in ('none', 'name'):
            if side == 'vertical':
                del_port = 2*del1*self.port_sep_y
            elif side == 'horizontal':
                del_port = 2*del1*self.port_sep_x
            if _flag1 == 'normal':
                _offset = (U - (N-1)*del_port - h1)/2 + k*del_port
            elif _flag1 == 'reverse':
                _offset = (U - (N-1)*del_port - h1)/2 + (N-k-1)*del_port
        else:
            if _flag1 == 'normal':
                _offset = (U - (N-1)*PORT_SEPARATION - h1)/2 + k*PORT_SEPARATION
            elif _flag1 == 'reverse':
                _offset = (U - (N-1)*PORT_SEPARATION - h1)/2 + (N-k-1)*PORT_SEPARATION

        return _offset

    def get_offset_1(self, side, port_offset, U, N, h1, k):

        _flag1 = 'normal'

        del1 = Constants.CANVAS_GRID_SIZE

        if side == 'vertical':
            del_port = 2*del1*self.port_sep_y
        elif side == 'horizontal':
            del_port = 2*del1*self.port_sep_x

        _offset = (U - (N-1)*del_port - h1)/2 + k*del_port + del1*port_offset

        return _offset

    def create_block_label(self, cr=None):
        """Create label for the block"""

        self.label_layout = Gtk.DrawingArea().create_pango_layout('')
        if cr:
            PangoCairo.update_layout(cr, self.label_layout)

    def create_shapes(self):
        """Update the block, parameters, and ports when a change occurs."""

        self.mirror = self.params['mirror'].get_value()
        self.rotate_strict = self.params['rotate_strict'].get_value()

        flag_ignore_mirror = False
        if not self.mirror == 'none':
            if self.rotate_strict == 'no':
                print('canvas/block.py: rotate_strict must be yes for mirroring.')
                print('   element:', self.name)
                print('   treating rotate_strict to be yes')
                self.rotate_strict = 'yes'
            if self.name.startswith('connector_'):
                print('canvas/block.py: mirroring not available for connector elements.')
                print('   element:', self.name)
                print('   mirror will be ignored.')
                flag_ignore_mirror = True

        if flag_ignore_mirror:
            self.mirror = 'none'

        if self.is_horizontal():
            self._area = (0, 0, self.width, self.height)
        elif self.is_vertical():
            self._area = (0, 0, self.height, self.width)

        self.bounds_from_area(self._area)

        if self.name.startswith('connector_'): self.mirror = 'none'

        if self.name.startswith('connector_f'):
            n_outputs = len(self.outputs_data)
            l_output_pos = []
            for k in range(n_outputs):
                pos = self.params['output' + str(k+1)].get_value()
                l_output_pos.append(pos)
            h1 = self.active_sources[0].height

            for k, port in enumerate(self.active_sources):
                port.create_shapes()
                pos = l_output_pos[k]

                if pos == 'right':
                    offset = self.get_offset('no', 'vertical', self.rotation, 0,
                       self.height, 1, h1, 0)

                    port.coordinate = {
                        0: (+self.width, offset),
                        90: (offset, -port.width),
                        180: (-port.width, offset),
                        270: (offset, +self.width),
                    }[port.connector_direction]
                else:
                    offset = self.get_offset(
                       'no', 'horizontal', self.rotation, 0,
                       self.width, 1, h1, 0)
                    port.coordinate = {
                        0: (+self.height, offset),
                        90: (offset, -port.width),
                        180: (-port.width, offset),
                        270: (offset, +self.height),
                    }[port.connector_direction]

        if self.mirror == 'none':
            if not self.name.startswith('connector_f'):
                if self.active_sources or self.e_rights or self.b_rights:
                    if self.active_sources:
                        h1 = self.active_sources[0].height
                    elif self.e_rights:
                        h1 = self.e_rights[0].height
                    elif self.b_rights:
                        h1 = self.b_rights[0].height

                    _N = len(self.active_sources) + len(self.e_rights) + len(self.b_rights)
                    _k = 0
                    for ports in (self.active_sources, self.e_rights, self.b_rights):
                        for port in ports:
                            port.create_shapes()

                            offset = self.get_offset(
                               self.params['rotate_strict'].get_value(),
                               'vertical', self.rotation, self.port_offset_r,
                               self.height, _N, h1, _k)

                            port.coordinate = {
                                0: (+self.width, offset),
                                90: (offset, -port.width),
                                180: (-port.width, offset),
                                270: (offset, +self.width),
                            }[port.connector_direction]
                            _k += 1

            if self.active_sinks or self.e_lefts or self.b_lefts:
                if self.active_sinks:
                    h1 = self.active_sinks[0].height
                elif self.e_lefts:
                    h1 = self.e_lefts[0].height
                elif self.b_lefts:
                    h1 = self.b_lefts[0].height

                _N = len(self.active_sinks) + len(self.e_lefts) + len(self.b_lefts)
                _k = 0
                for ports in (self.active_sinks, self.e_lefts, self.b_lefts):
                    for port in ports:
                        port.create_shapes()
                        offset = self.get_offset(
                           self.params['rotate_strict'].get_value(),
                           'vertical', self.rotation, self.port_offset_l,
                           self.height, _N, h1, _k)
                        port.coordinate = {
                            0: (+self.width, offset),
                            90: (offset, -port.width),
                            180: (-port.width, offset),
                            270: (offset, +self.width),
                        }[port.connector_direction]
                        _k += 1

            if self.e_tops or self.b_tops:
                if self.e_tops:
                    h1 = self.e_tops[0].height
                elif self.b_tops:
                    h1 = self.b_tops[0].height

                _N = len(self.e_tops) + len(self.b_tops)
                _k = 0
                for ports in (self.e_tops, self.b_tops):
                    for port in ports:
                        port.create_shapes()
                        offset = self.get_offset(
                           self.params['rotate_strict'].get_value(),
                           'horizontal', self.rotation, self.port_offset_t,
                           self.width, _N, h1, _k)
                        port.coordinate = {
                            0: (+self.height, offset),
                            90: (offset, -port.width),
                            180: (-port.width, offset),
                            270: (offset, +self.height),
                        }[port.connector_direction]
                        _k += 1

            if self.e_bottoms or self.b_bottoms:
                if self.e_bottoms:
                    h1 = self.e_bottoms[0].height
                elif self.b_bottoms:
                    h1 = self.b_bottoms[0].height

                _N = len(self.e_bottoms) + len(self.b_bottoms)
                _k = 0
                for ports in (self.e_bottoms, self.b_bottoms):
                    for port in ports:
                        port.create_shapes()
                        offset = self.get_offset(
                           self.params['rotate_strict'].get_value(),
                           'horizontal', self.rotation, self.port_offset_b,
                           self.width, _N, h1, _k)
                        port.coordinate = {
                            0: (+self.height, offset),
                            90: (offset, -port.width),
                            180: (-port.width, offset),
                            270: (offset, +self.height),
                        }[port.connector_direction]
                        _k += 1

            if self.drawing_scheme == 'symbol':
                rotate1 = Utils.get_rotated_coordinate_1

                self.c__ = []
                for c1 in self.c_:
                    self.c__.append(rotate1(c1, self.width, self.height, self.rotation))

                self.a__ = []
                for theta in self.a_: 
                    self.a__.append(Utils.get_rotated_angle_1(theta, self.rotation))

                if self.s_[0]:
                    self.s__ = Utils.get_scaling_factors(self.s_, self.rotation)

        else:

            flag_v = True if self.mirror == 'v' else False
            flag_h = True if self.mirror == 'h' else False

            Hb, Wb = self.height, self.width

            if self.active_sources or self.e_rights or self.b_rights:
                if self.active_sources:
                    h1 = self.active_sources[0].height
                elif self.e_rights:
                    h1 = self.e_rights[0].height
                elif self.b_rights:
                    h1 = self.b_rights[0].height

                _N = len(self.active_sources) + len(self.e_rights) + len(self.b_rights)
                _k = 0
                for ports in (self.active_sources, self.e_rights, self.b_rights):
                    for port in ports:
                        port.create_shapes_1()
                        Hp, Wp = port.height, port.width

                        offset = self.get_offset_1('vertical',
                            self.port_offset_r, Hb, _N, h1, _k)

                        port.coordinate = {
                            (  0, 'v'): (-Wp         , offset      ),
                            (180, 'h'): (-Wp         , offset      ),
                            ( 90, 'v'): (offset      , Wb          ),
                            (270, 'h'): (offset      , Wb          ),
                            (180, 'v'): (Wb          , Hb-offset-Hp),
                            (  0, 'h'): (Wb          , Hb-offset-Hp),
                            (270, 'v'): (Hb-offset-Hp, -Wp         ),
                            ( 90, 'h'): (Hb-offset-Hp, -Wp         ),
                        }[(self.rotation, self.mirror)]

                        _k += 1
            if self.active_sinks or self.e_lefts or self.b_lefts:
                if self.active_sinks:
                    h1 = self.active_sinks[0].height
                elif self.e_lefts:
                    h1 = self.e_lefts[0].height
                elif self.b_lefts:
                    h1 = self.b_lefts[0].height

                _N = len(self.active_sinks) + len(self.e_lefts) + len(self.b_lefts)
                _k = 0
                for ports in (self.active_sinks, self.e_lefts, self.b_lefts):
                    for port in ports:
                        port.create_shapes_1()
                        Hp, Wp = port.height, port.width

                        offset = self.get_offset_1('vertical',
                            self.port_offset_l, Hb, _N, h1, _k)

                        port.coordinate = {
                            (  0, 'v'): (Wb          , offset      ),
                            (180, 'h'): (Wb          , offset      ),
                            ( 90, 'v'): (offset      , -Wp         ),
                            (270, 'h'): (offset      , -Wp         ),
                            (180, 'v'): (-Wp         , Hb-offset-Hp),
                            (  0, 'h'): (-Wp         , Hb-offset-Hp),
                            (270, 'v'): (Hb-offset-Hp, Wb          ),
                            ( 90, 'h'): (Hb-offset-Hp, Wb          ),
                        }[(self.rotation, self.mirror)]

                        _k += 1

            if self.e_tops or self.b_tops:
                if self.e_tops:
                    h1 = self.e_tops[0].height
                elif self.b_tops:
                    h1 = self.b_tops[0].height

                _N = len(self.e_tops) + len(self.b_tops)
                _k = 0
                for ports in (self.e_tops, self.b_tops):
                    for port in ports:
                        port.create_shapes_1()
                        Hp, Wp = port.height, port.width

                        offset = self.get_offset_1('horizontal',
                            self.port_offset_t, Wb, _N, h1, _k)
                        port.coordinate = {
                            (  0, 'v'): (Wb-offset-Hp, -Wp         ),
                            (180, 'h'): (Wb-offset-Hp, -Wp         ),
                            ( 90, 'v'): (-Wp         , offset      ),
                            (270, 'h'): (-Wp         , offset      ),
                            (180, 'v'): (offset      , Hb          ),
                            (  0, 'h'): (offset      , Hb          ),
                            (270, 'v'): (Hb          , Wb-offset-Hp),
                            ( 90, 'h'): (Hb          , Wb-offset-Hp),
                        }[(self.rotation, self.mirror)]
                        _k += 1

            if self.e_bottoms or self.b_bottoms:
                if self.e_bottoms:
                    h1 = self.e_bottoms[0].height
                elif self.b_bottoms:
                    h1 = self.b_bottoms[0].height

                _N = len(self.e_bottoms) + len(self.b_bottoms)
                _k = 0
                for ports in (self.e_bottoms, self.b_bottoms):
                    for port in ports:
                        port.create_shapes_1()
                        Hp, Wp = port.height, port.width

                        offset = self.get_offset_1('horizontal',
                            self.port_offset_b, Wb, _N, h1, _k)
                        port.coordinate = {
                            (  0, 'v'): (Wb-offset-Hp, Hb          ),
                            (180, 'h'): (Wb-offset-Hp, Hb          ),
                            ( 90, 'v'): (Hb          , offset      ),
                            (270, 'h'): (Hb          , offset      ),
                            (180, 'v'): (offset      , -Wp         ),
                            (  0, 'h'): (offset      , -Wp         ),
                            (270, 'v'): (-Wp         , Wb-offset-Hp),
                            ( 90, 'h'): (-Wp         , Wb-offset-Hp),
                        }[(self.rotation, self.mirror)]
                        _k += 1

            if self.drawing_scheme == 'symbol':
                rotate2 = Utils.get_rotated_coordinate_2

                self.c__ = []
                for c1 in self.c_:
                    self.c__.append(rotate2(c1, self.width, self.height, self.rotation, self.mirror))

                self.a__ = []
                for theta in self.a_: 
                    self.a__.append(Utils.get_rotated_angle_2(theta, self.rotation, self.mirror))

                if self.s_[0]:
                    self.s__ = Utils.get_scaling_factors(self.s_, self.rotation)

    def create_labels(self, cr=None):
        """Create the labels for the signal block."""

        self.create_block_label()

        def options_create_markup(s1, s2):
            s_font = Constants.PARAM_FONT
            s3 = '<span font_desc=' + '"' + s_font + '">' \
               + '<b>{label}:</b> {value}</span>'.format(label=Utils.encode(s1), value=s2)
            return s3

        def create_markup_1(s1):
            s_font = Constants.BLOCK_FONT
            s3 = '<span font_desc=' + '"' + s_font + '">' \
               + '<b>{label}</b></span>'.format(label=Utils.encode(s1))
            return s3

        def get_min_dim(ports):

            n_ports = len(ports)
            if n_ports > 1:
                min_dim = 2*PORT_BORDER_SEPARATION + (n_ports-1)*PORT_SEPARATION
            else:
                min_dim = 2*PORT_BORDER_SEPARATION

            return min_dim

        if self.drawing_scheme == 'name':
            # (Re-)creating layouts here, because layout.context_changed() doesn't seems to work (after zoom)
            title_layout, params_layout = self._surface_layouts = [
                Gtk.DrawingArea().create_pango_layout(''),  # title
                Gtk.DrawingArea().create_pango_layout(''),  # params
            ]

            if cr:  # to fix up extents after zooming
                PangoCairo.update_layout(cr, title_layout)
                PangoCairo.update_layout(cr, params_layout)

            foreground='foreground="red"' if not self.is_valid() else ''
            label=Utils.encode(self.label)

            block_type = type(self).__name__
            if block_type == 'dummy_source':
                s1 = self.params['n'].get_value() + ' >'
            elif block_type == 'dummy_sink':
                s1 = '> ' + self.params['n'].get_value()
            elif block_type.startswith('dummy_e'):
                s1 = self.params['n'].get_value()
            elif block_type.startswith('dummy_b'):
                s1 = self.params['n'].get_value()
            elif block_type.startswith('pad_source'):
                s1 = self.params['label'].get_value() + ' >'
            elif block_type.startswith('pad_sink'):
                s1 = '> ' + self.params['label'].get_value()
            elif block_type.startswith('pad_e_'):
                s1 = self.params['label'].get_value()
            elif block_type.startswith('pad_b_'):
                s1 = self.params['label'].get_value()
            else:
                s2 = self.label
                s1 = s2.split('\\n')[0] if '\\n' in s2 else s2

            if block_type.startswith('dummy_'):
                title_layout.set_markup(
                    '<span {foreground} font_desc="{font}">{label}</span>'.format(
                        foreground='foreground="red"' if not self.is_valid() else '', font=BLOCK_FONT,
                        label=Utils.encode(s1)
                    )
                )
            else:
                title_layout.set_markup(
                    '<span {foreground} font_desc="{font}"><b>{label}</b></span>'.format(
                        foreground='foreground="red"' if not self.is_valid() else '', font=BLOCK_FONT,
                        label=Utils.encode(s1)
                    )
                )

            title_width, title_height = title_layout.get_size()
 
            force_show_id = False

            if type(self).__name__ == 'options':

                markups = []
                l = [
                   'title',
                   'wiring_style',
                   'engine_output',
                   'generate_options',
                ]
                for param in self.params.values():
                    if param.key in l:
                        s1 = param.name
                        s2 = self.params[param.key].get_value()
                        s3 = options_create_markup(s1, s2)
                        markups.append(s3)

            else:
                markups = []
                s1 = self.label
                if '\\n' in s1:
                    l = s1.split('\\n')
                    markups = []
                    for k in range(1, len(l)):
                        s3 = create_markup_1(l[k])
                        markups.append(s3)

            params_layout.set_spacing(LABEL_SEPARATION * Pango.SCALE)
            params_layout.set_markup('\n'.join(markups))
            params_width, params_height = params_layout.get_size() if markups else (0, 0)

            label_width = max(title_width, params_width) / Pango.SCALE
            label_height = title_height / Pango.SCALE
            if markups:
                label_height += LABEL_SEPARATION + params_height / Pango.SCALE

            if self.name.startswith('dummy_'):
                width = label_width + 2 * BLOCK_LABEL_PADDING_TAGS
                height = label_height + 2 * BLOCK_LABEL_PADDING_TAGS
            else:
                width = label_width + 2 * BLOCK_LABEL_PADDING
                height = label_height + 2 * BLOCK_LABEL_PADDING

            self._update_colors()
            self.create_port_labels()

#           Note: for xbe/ebe blocks, we will not have both x and e nodes on the
#              same edge, but we need to keep this for subckts:

            if self.name.startswith('dummy_'):
                min_h_1 = get_min_dim(self.active_sinks + self.e_lefts)
                min_h_2 = get_min_dim(self.active_sources + self.e_rights)
                height = max(height, min_h_1, min_h_2)
                min_w_1 = get_min_dim(self.e_tops)
                min_w_2 = get_min_dim(self.e_bottoms)
                width = max(width, min_w_1, min_w_2)
            else:
                if self.name.startswith('connector_f'):
                    n_ports_W = n_ports_E = n_ports_S = n_ports_N = 1
                else:

                    n_ports_W = len(self.active_sinks) + len(self.e_lefts) + len(self.b_lefts)
                    n_ports_E = len(self.active_sources) + len(self.e_rights) + len(self.b_rights)
                    n_ports_N = len(self.e_tops) + len(self.b_tops)
                    n_ports_S = len(self.e_bottoms) + len(self.b_bottoms)

                n_ports_WE = max(n_ports_W, n_ports_E)
                n_ports_SN = max(n_ports_S, n_ports_N)

                del1 = Constants.CANVAS_GRID_SIZE

                if n_ports_WE > 0:
                    height1 = 2*del1*(self.port_sep_y*(n_ports_WE-1) + self.port_block_y)
                else:
                    height1 = 2*del1*self.port_block_y

                if n_ports_SN > 0:
                    width1 = 2*del1*(self.port_sep_x*(n_ports_SN-1) + self.port_block_x)
                else:
                    width1 = 2*del1*self.port_block_x

                height = max(height, height1)
                width = max(width, width1)

            self.width, self.height = width, height = Utils.align_to_grid_1((width, height))

            self._surface_layouts_offsets = [
                (0, (height - label_height) / 2.0),
                (0, (height - label_height) / 2.0 + LABEL_SEPARATION + title_height / Pango.SCALE)
            ]

            title_layout.set_width(width * Pango.SCALE)
            title_layout.set_alignment(Pango.Alignment.CENTER)
            params_layout.set_indent((width - label_width) / 2.0 * Pango.SCALE)

        elif self.drawing_scheme == 'text':

            title_layout, params_layout = self._surface_layouts = [
                Gtk.DrawingArea().create_pango_layout(''),
                Gtk.DrawingArea().create_pango_layout(''),
            ]
            if cr:
                PangoCairo.update_layout(cr, title_layout)
                PangoCairo.update_layout(cr, params_layout)

            foreground = ''
            label=Utils.encode(self.label)

            if self.key == 'show_text':

                font_size = self.params['font_size'].get_value()
                s_font = "Sans " + font_size

                s1 = self.params['text'].get_value()
                l1 = [x.strip() for x in s1.split('\\n')]
                s_title = l1[0] if l1 else 'dummy'
                markups = []
                for i, s in enumerate(l1):
                    if i > 0:
                        s2 = '<span font_desc=' + '"' + s_font + '">' + s + '</span>'
                        markups.append(s2)

            elif self.key == 'show_parameter':
                s1 = self.params['element_name'].get_value()
                s2 = self.params['parameter_name'].get_value()

                l3 = [x for x in self.parent0.blocks if x.name == s1]
                flag_found = False

                if l3:
                    b1 = l3[0]

                    if s2 in b1.params.keys():
                        s_param = b1.params[s2].get_value()
                        flag_found = True
                        flag_show_name = self.params['show_name'].get_value() == 'show'

                if flag_found:
                    if flag_show_name:
                        s_title = s2 + '=' + s_param
                    else:
                        s_title = s_param
                else:
                    s_title = 'error'

                font_size = self.params['font_size'].get_value()
                s_font = "Sans " + font_size
                markups = []

            title_layout.set_markup('<span font_desc="{font}">{name}</span>'.format(
                name=Utils.encode(s_title), font=s_font))

            title_width, title_height = title_layout.get_size()

            params_layout.set_spacing(SHOW_TEXT_LABEL_SEPARATION * Pango.SCALE)
            params_layout.set_markup('\n'.join(markups))
            params_width, params_height = params_layout.get_size() if markups else (0, 0)

            label_width = max(title_width, params_width) / Pango.SCALE
            label_height = title_height / Pango.SCALE
            if markups:
                label_height += SHOW_TEXT_LABEL_SEPARATION + params_height / Pango.SCALE

            width = label_width + 2 * TEXT_BLOCK_LABEL_PADDING
            height = label_height + 2 * TEXT_BLOCK_LABEL_PADDING

            self._update_colors()

            self.width, self.height = width, height = Utils.align_to_grid_1((width, height))

            self._surface_layouts_offsets = [
                (0, (height - label_height) / 2.0),
                (0, (height - label_height) / 2.0 + SHOW_TEXT_LABEL_SEPARATION + title_height / Pango.SCALE)
            ]

            title_layout.set_width(width * Pango.SCALE)
            title_layout.set_alignment(Pango.Alignment.LEFT)
            params_layout.set_alignment(Pango.Alignment.LEFT)
            params_layout.set_indent((width - label_width) / 2.0 * Pango.SCALE)

        elif self.drawing_scheme in ('none', 'symbol'):
            self._update_colors()
            self.create_port_labels()

            if self.name.startswith('connector_f'):
                n_ports_W = n_ports_E = n_ports_S = n_ports_N = 1
            else:

                n_ports_W = len(self.active_sinks) + len(self.e_lefts) + len(self.b_lefts)
                n_ports_E = len(self.active_sources) + len(self.e_rights) + len(self.b_rights)
                n_ports_N = len(self.e_tops) + len(self.b_tops)
                n_ports_S = len(self.e_bottoms) + len(self.b_bottoms)

            n_ports_WE = max(n_ports_W, n_ports_E)
            n_ports_SN = max(n_ports_S, n_ports_N)

            del1 = Constants.CANVAS_GRID_SIZE

            if n_ports_WE > 0:
                self.height = 2*del1*(self.port_sep_y*(n_ports_WE-1) + self.port_block_y)
            else:
                self.height = 2*del1*self.port_block_y

            if n_ports_SN > 0:
                self.width = 2*del1*(self.port_sep_x*(n_ports_SN-1) + self.port_block_x)
            else:
                self.width = 2*del1*self.port_block_x

    def create_port_labels(self):
        for ports in (self.active_sinks, self.active_sources,
            self.e_lefts, self.e_rights, self.e_tops, self.e_bottoms,
            self.b_lefts, self.b_rights, self.b_tops, self.b_bottoms):
            max_width = 0
            for port in ports:
                port.create_labels()
                max_width = max(max_width, port.width_with_label)
            for port in ports:
                port.width = max_width

    def draw(self, cr):
        """
        Draw the signal block with label and inputs/outputs.
        """

        if self.key == 'options':

            delta1 = int(self.params['delta_show_grid'].get_value())

            if delta1 != 0:

                w0a, h0a = self.parent._main_window.get_size()

                w0 = max(w0a, self.parent.max_w_all + 50)
                h0 = max(h0a, self.parent.max_h_all + 50)

                d1_grid = delta1*RECT_WIRING_DELTA1

                x_l = 0
                x_r = w0
                y_t = 0
                y_b = h0

                cr.save()
                cr.set_line_width(0.5)

                nttlx = 0
                httl = 0

                while httl < h0:
                    if (nttlx % 5) == 0:
                        cr.set_source_rgba(*colors.GRID_DARK_COLOR)
                    else:
                        cr.set_source_rgba(*colors.GRID_LIGHT_COLOR)
                    cr.move_to(x_l, httl)
                    cr.line_to(x_r, httl)
                    cr.stroke()
                    httl += d1_grid
                    nttlx += 1

                nttly = 0
                wttl = 0

                while wttl < w0:
                    if (nttly % 5) == 0:
                        cr.set_source_rgba(*colors.GRID_DARK_COLOR)
                    else:
                        cr.set_source_rgba(*colors.GRID_LIGHT_COLOR)
                    cr.move_to(wttl, y_t)
                    cr.line_to(wttl, y_b)
                    cr.stroke()
                    wttl += d1_grid
                    nttly += 1

                cr.restore()

        cr.translate(*self.coordinate)

        if self._show_label:

            s_font = "Sans " + str(Constants.BLOCK_LABEL_FONTSIZE)

            self.label_layout.set_markup('<span font_desc="{font}">{name}</span>'.format(
                name=Utils.encode(self.name), font=s_font))

            self.label_layout.set_alignment(Pango.Alignment.LEFT)

            cr.translate(0, -Constants.BLOCK_LABEL_OFFSET)
            cr.set_source_rgba(*colors.BLOCK_LABEL_COLOR)

            PangoCairo.update_layout(cr, self.label_layout)
            PangoCairo.show_layout(cr, self.label_layout)

            cr.translate(0, Constants.BLOCK_LABEL_OFFSET)

        if self.drawing_scheme == 'name':
            border_color = colors.HIGHLIGHT_COLOR if self.highlighted else self._border_color

            for port in self.active_ports():  # ports first
                cr.save()
                port.draw(cr)
                cr.restore()

            cr.rectangle(*self._area)
            cr.set_source_rgba(*self._bg_color)
            cr.fill_preserve()
            cr.set_source_rgba(*border_color)

            cr.set_line_width(0.5)

            cr.stroke()

            # title and params label
            if self.is_vertical():
                cr.rotate(-math.pi / 2)
                cr.translate(-self.width, 0)
            cr.set_source_rgba(*self._font_color)

            for layout, offset in zip(self._surface_layouts, self._surface_layouts_offsets):
                cr.save()
                cr.translate(*offset)
                PangoCairo.update_layout(cr, layout)
                PangoCairo.show_layout(cr, layout)
                cr.restore()

        elif self.drawing_scheme == 'text':
            border_color = colors.HIGHLIGHT_COLOR if self.highlighted else colors.FLOWGRAPH_BACKGROUND_COLOR

            cr.rectangle(*self._area)
            cr.set_source_rgba(*colors.FLOWGRAPH_BACKGROUND_COLOR)
            cr.fill_preserve()
            cr.set_source_rgba(*border_color)

            cr.set_line_width(0.5)
            cr.stroke()

            if self.is_vertical():
                cr.rotate(-math.pi / 2)
                cr.translate(-self.width, 0)
            cr.set_source_rgba(*colors.SHOW_TEXT_COLOR)

            for layout, offset in zip(self._surface_layouts, self._surface_layouts_offsets):
                cr.save()
                cr.translate(*offset)
                PangoCairo.update_layout(cr, layout)
                PangoCairo.show_layout(cr, layout)
                cr.restore()

        elif self.drawing_scheme == 'none':
            border_color = colors.HIGHLIGHT_COLOR if self.highlighted else self._border_color

            for port in self.active_ports():  # ports first
                cr.save()
                port.draw(cr)
                cr.restore()

            cr.rectangle(*self._area)
            cr.set_source_rgba(*self._bg_color)
            cr.fill_preserve()
            cr.set_source_rgba(*border_color)

            cr.set_line_width(0.5)

            cr.stroke()
        elif self.drawing_scheme == 'symbol':

            for port in self.active_ports():  # ports first
                cr.save()
                port.draw(cr)
                cr.restore()

#           if self.highlighted:
#               border_color = colors.HIGHLIGHT_COLOR
#               cr.rectangle(*self._area)
#           else:
#               border_color = colors.FLOWGRAPH_BACKGROUND_COLOR
#               x1, y1, x2, y2 = self._area
#               cr.rectangle(x1+2, y1+2, x2-5, y2-5)
#           cr.set_source_rgba(*colors.FLOWGRAPH_BACKGROUND_COLOR)
#           cr.fill_preserve()
#           cr.set_line_width(1.0)
#           cr.set_source_rgba(*border_color)
#           cr.stroke()

            border_color = colors.FLOWGRAPH_BACKGROUND_COLOR
            x1, y1, x2, y2 = self._area
            cr.rectangle(x1+2, y1+2, x2-5, y2-5)
            cr.set_source_rgba(*colors.FLOWGRAPH_BACKGROUND_COLOR)
            cr.fill_preserve()
            cr.set_line_width(1.0)
            cr.set_source_rgba(*border_color)
            cr.stroke()

            cr.set_source_rgb(0, 0, 0)

            if self.key in('connector_e_2A', 'connector_b_2A', 'connector_f_2A', 'connector_f_2B'):
                cr.set_line_width(Constants.temp1)
            else:
                cr.set_line_width(1.5*Constants.temp1)

            cr.set_line_join(cairo.LineJoin.ROUND)

            c_ = self.c__
            a_ = self.a__
            s_ = self.s__
            t_ = self.t_

            if self.has_arc:
                if self.mirror == 'none':
                    for s in self.l_draw: exec(s)
                else:
                    l_draw_new = []
                    for s in self.l_draw:
                        if 'cr.arc' in s:
                            s1 = s.replace('cr.arc(', 'cr.dummy1')
                            s2 = s1.replace('cr.arc_negative(', 'cr.arc(')
                            s3 = s2.replace('cr.dummy1', 'cr.arc_negative(')
                        else:
                            s3 = s
                        l_draw_new.append(s3)

                    for s in l_draw_new: exec(s)
            else:
                for s in self.l_draw: exec(s)

            if self.highlighted:
                border_color = colors.HIGHLIGHT_COLOR
                cr.rectangle(*self._area)
                cr.set_line_width(1.0)
                cr.set_source_rgba(*border_color)
                cr.stroke()

    def what_is_selected(self, coor, coor_m=None):
        """
        Get the element that is selected.

        Args:
            coor: the (x,y) tuple
            coor_m: the (x_m, y_m) tuple

        Returns:
            this block, a port, or None
        """
        for port in self.active_ports():
            port_selected = port.what_is_selected(
                coor=[a - b for a, b in zip(coor, self.coordinate)],
                coor_m=[a - b for a, b in zip(coor, self.coordinate)] if coor_m is not None else None
            )
            if port_selected:
                return port_selected
        return Drawable.what_is_selected(self, coor, coor_m)

    def draw_comment(self, cr):
        if not self._comment_layout:
            return
        x, y = self.coordinate

        if self.is_horizontal():
            y += self.height + BLOCK_LABEL_PADDING
        else:
            x += self.height + BLOCK_LABEL_PADDING

        cr.save()
        cr.translate(x, y)
        PangoCairo.update_layout(cr, self._comment_layout)
        PangoCairo.show_layout(cr, self._comment_layout)
        cr.restore()

    def get_extents(self):
        extent = Drawable.get_extents(self)
        x, y = self.coordinate
        for port in self.active_ports():
            extent = (min_or_max(xy, offset + p_xy) for offset, min_or_max, xy, p_xy in zip(
                (x, y, x, y), (min, min, max, max), extent, port.get_extents()
            ))
        return tuple(extent)

    def get_extents_comment(self):
        x, y = self.coordinate
        if not self._comment_layout:
            return x, y, x, y
        if self.is_horizontal():
            y += self.height + BLOCK_LABEL_PADDING
        else:
            x += self.height + BLOCK_LABEL_PADDING
        w, h = self._comment_layout.get_pixel_size()
        return x, y, x + w, y + h

    def mouse_over(self):
        changed = not self._show_label
        self._hovering = True
        return changed

    def mouse_out(self):
        label_was_shown = self._show_label
        self._hovering = False
        return label_was_shown != self._show_label

    @property
    def _show_label(self):
        return self._hovering




# Copyright 2008-2017 Free Software Foundation, Inc.
# This file is part of GNU Radio
#
# GNU Radio Companion is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# GNU Radio Companion is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

from __future__ import absolute_import

import re

from .. import blocks
from .. import Constants

# Blacklist certain ids, its not complete, but should help

ID_BLACKLIST = ['self', 'options', 'gr', 'math', 'firdes']

validators = {}

def validates(*dtypes):
    def decorator(func):
        for dtype in dtypes:
            assert dtype in Constants.PARAM_TYPE_NAMES
            validators[dtype] = func
        return func
    return decorator

class ValidateError(Exception):
    """Raised by validate functions"""

@validates('id')
def validate_block_id(param):
    value = param.value
    # Can python use this as a variable?
    if not re.match(r'^[a-z|A-Z|_]*[\$0-9]*', value):
        raise ValidateError('ID "{}" must begin with a letter and may contain letters, numbers, '
                            'and underscores.'.format(value))
    if value in ID_BLACKLIST:
        raise ValidateError('ID "{}" is blacklisted.'.format(value))
    block_names = [block.name for block in param.parent_flowgraph.iter_enabled_blocks()]
    # Id should only appear once, or zero times if block is disabled
    if param.key == 'id' and block_names.count(value) > 1:
        raise ValidateError('ID "{}" is not unique.'.format(value))
    elif value not in block_names:
        raise ValidateError('ID "{}" does not exist.'.format(value))
    return value

@validates('name')
def validate_name(param):
    # Name of a function that will be generated literally not as a string
    value = param.value
    # Can python use this as a variable?
    if not re.match(r'^[a-z|A-Z|_]*[\$0-9]*', value):
        raise ValidateError('ID "{}" must begin with a letter and may contain letters, numbers, '
                            'and underscores.'.format(value))
    return value




"""
Copyright (C) 2022 - Mahesh Patil <mbpatil@ee.iitb.ac.in>
This file is part of GSEIM.

GSEIM is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk

class DisplayName(Gtk.Dialog):
    def __init__(self, parent, s_name, l1):
        Gtk.Dialog.__init__(self, title="Element Name", transient_for=parent, flags=0)
        self.add_buttons(
            Gtk.STOCK_OK, Gtk.ResponseType.OK
        )
        self.set_default_size(250, 100)
        self.set_position(Gtk.WindowPosition.CENTER)

        box = self.get_content_area()
        box.set_spacing(30)
        box.set_margin_top(5)

        label = Gtk.Label(s_name, xalign=0.5)
        box.add(label)

        if l1:
            for l2 in l1:
               label = Gtk.Label(l2[0] + ': ' + l2[1], xalign=0)
               box.add(label)

        self.show_all()



from setuptools import setup, find_packages


setup(
    name='GSEIMIITB',
    version='1.0',
    # license='MIT',
    # author="Piyush Palav",
    # author_email='piyushdpalav@gmail.com',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    url='https://github.com/PiyushPalav/gseim',
    keywords='gseim',
    install_requires=[
          'numpy',
          'matplotlib',
          'psutil',
          'pyyaml'
      ],
)



import codecs
import operator
import os
import tempfile
import textwrap

from .. import Messages, blocks
from .FlowGraphProxy import FlowGraphProxy
from ..utils import expr_utils

DATA_DIR = os.path.dirname(__file__)

class TopBlockGenerator(object):

    def __init__(self, flow_graph, file_path):
        """
        Initialize the top block generator object.

        Args:
            flow_graph: the flow graph object
            file_path: the path to write the file to
        """

        self._flow_graph = FlowGraphProxy(flow_graph)
        self._generate_options = self._flow_graph.get_option('generate_options')

        # Handle the case where the directory is read-only
        # In this case, use the system's temp directory
        if not os.access(file_path, os.W_OK):
            file_path = tempfile.gettempdir()
        filename = self._flow_graph.get_option('id') + '.py'
        self.file_path = os.path.join(file_path, filename)
        self._dirname = file_path
        print("TopBlockGenerator: filename =",filename)
        print("TopBlockGenerator: file_path =",file_path)

    def write(self):
        """generate output and write it to files"""

        fg = self._flow_graph
        self.title = fg.get_option('title') or fg.get_option('id').replace('_', ' ').title()
        variables = fg.get_variables()
        parameters = fg.get_parameters()

        self.namespace = {
            'flow_graph': fg,
            'variables': variables,
            'parameters': parameters,
            'generate_options': self._generate_options,
        }




"""
Copyright 2007, 2008, 2009 Free Software Foundation, Inc.
This file is part of GNU Radio

GNU Radio Companion is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

GNU Radio Companion is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""

from __future__ import absolute_import, division

import math

from gi.repository import Gtk, PangoCairo, Pango

from . import colors
from .drawable import Drawable
from .. import Actions, Utils, Constants

from ...core.utils.descriptors import nop_write
from ...core.ports import Port as CorePort

class Port(CorePort, Drawable):
    """The graphical port."""

    def __init__(self, parent, direction, **n):
        """
        Port constructor.
        Create list of connector coordinates.
        """
        super(self.__class__, self).__init__(parent, direction, **n)
        Drawable.__init__(self)
        self._connector_coordinate = (0, 0)
        self._hovering = False
        self.force_show_label = False

        self._area = []
        self._bg_color = self._border_color = 0, 0, 0, 0
        self._font_color = list(colors.PORT_NAME_COLOR)

        self._line_width_factor = 1.0
        self._label_layout_offsets = 0, 0

        self.width_with_label = self.height = 0

        self.label_layout = None

        self.port_category = direction

        self.port_index = n['id']

    @property
    def width(self):
#       return self.width_with_label if self._show_label else Constants.PORT_LABEL_HIDDEN_WIDTH
        return Constants.PORT_DIM

    @width.setter
    def width(self, value):
        self.width_with_label = value
        self.label_layout.set_width(value * Pango.SCALE)

    def create_shapes(self):
        """Create new areas and labels for the port."""

#       Note: is_horizontal indicates rotation of the block -> even though
#       the top and bottom nodes are different in orientation than the left
#       and right nodes, we will allocate the same rotation value to all of
#       them, i.e., all of them would be horizontal or all of them would be
#       vertical.

        if self.port_category in ('e_top', 'b_top', 'e_bottom', 'b_bottom'):
            if self.is_horizontal():
                self._area = (0, 0, self.height, self.width)
            elif self.is_vertical():
                self._area = (0, 0, self.width, self.height)
        else:
            if self.parent_block.name.startswith('connector_f'):
                if self.is_source:
                    index1 = self.port_index
                    pos = self.parent_block.params['output' + str(int(index1)+1)].get_value()

                    if pos == 'top' or pos == 'bottom':
                        if self.is_horizontal():
                            self._area = (0, 0, self.height, self.width)
                        elif self.is_vertical():
                            self._area = (0, 0, self.width, self.height)
                    elif pos == 'right':
                        if self.is_horizontal():
                            self._area = (0, 0, self.width, self.height)
                        elif self.is_vertical():
                            self._area = (0, 0, self.height, self.width)
                else:
                    if self.is_horizontal():
                        self._area = (0, 0, self.width, self.height)
                    elif self.is_vertical():
                        self._area = (0, 0, self.height, self.width)
            else:
                if self.is_horizontal():
                    self._area = (0, 0, self.width, self.height)
                elif self.is_vertical():
                    self._area = (0, 0, self.height, self.width)

        if self.parent_block.name.startswith('connector_'):
            l = self._area
        else:
            delta = Constants.PORT_HOVER_DELTA
            l = [x - delta for x in self._area[:2]] + [x + 2*delta for x in self._area[2:]]

        self.bounds_from_area(l)

        self._connector_coordinate = {
            0:   (self.width, self.height / 2),
            90:  (self.height / 2, 0),
            180: (0, self.height / 2),
            270: (self.height / 2, self.width)
        }[self.connector_direction]

    def create_shapes_1(self):

        if self.port_category in ('e_top', 'b_top', 'e_bottom', 'b_bottom'):
            if self.is_horizontal():
                self._area = (0, 0, self.height, self.width)
            elif self.is_vertical():
                self._area = (0, 0, self.width, self.height)
        else:
            if self.is_horizontal():
                self._area = (0, 0, self.width, self.height)
            elif self.is_vertical():
                self._area = (0, 0, self.height, self.width)

        delta = Constants.PORT_HOVER_DELTA
        l = [x - delta for x in self._area[:2]] + [x + 2*delta for x in self._area[2:]]

        self.bounds_from_area(l)

        self._connector_coordinate = {
              0: (self.width, self.height / 2),
             90: (self.height / 2, 0),
            180: (0, self.height / 2),
            270: (self.height / 2, self.width)
        }[self.connector_direction]

    def create_labels(self, cr=None):
        """Create the labels for the socket."""
        self.label_layout = Gtk.DrawingArea().create_pango_layout('')

        if cr:
            PangoCairo.update_layout(cr, self.label_layout)

        self._line_width_factor = 1.0

        layout = self.label_layout

#       self.width = Constants.PORT_LABEL_HIDDEN_WIDTH
        self.width = Constants.PORT_DIM

#       self.height = 13
        self.height = Constants.PORT_DIM

        self.height += self.height % 2  # uneven height

    def draw(self, cr):
        """
        Draw the socket with a label.
        """

        border_color = colors.BORDER_COLOR

        cr.set_line_width(0.5)

        cr.translate(*self.coordinate)

#       this draws only the rectangles (fill is done below)
        if not self.parent_block.name.startswith('connector_'):
            cr.rectangle(*self._area)

        if self.parent_block.name.startswith('connector_'):
            cr.set_source_rgba(*colors.FLOWGRAPH_BACKGROUND_COLOR)

            _color1 = colors.CONNECTION_ENABLED_COLOR
            cr.move_to(*self._connector_coordinate)
            x, y = self._connector_coordinate

            if self.connector_direction == 0:
                x -= self.width
            elif self.connector_direction == 180:
                x += self.width
            elif self.connector_direction == 90:
                y += self.width
            elif self.connector_direction == 270:
                y -= self.width

            cr.set_line_width(Constants.temp1)
            cr.line_to(x, y)
        else:
            if self.port_category in ('e_left', 'e_right', 'e_top', 'e_bottom'):
                cr.set_source_rgba(*colors.E_PORT_COLOR)
            elif self.port_category in ('b_left', 'b_right', 'b_top', 'b_bottom'):
                cr.set_source_rgba(*colors.B_PORT_COLOR)
            else:
                cr.set_source_rgba(*colors.F_PORT_COLOR)

        cr.fill_preserve()
        cr.set_source_rgba(*border_color)

#       commenting this out removes the fill (but the ports are still drawn)
        cr.stroke()

#       debug
#       return
        if not self._show_label:
            return  # this port is folded (no label)

        if self.parent_block.name.startswith('connector_f'):
            if self.is_source:
                index1 = self.port_index
                pos = self.parent_block.params['output' + str(int(index1)+1)].get_value()
                if pos == 'top' or pos == 'bottom':
                    if self.is_horizontal():
                        cr.rotate(-math.pi / 2)
                        cr.translate(-self.width, 0)
                elif pos == 'right':
                    if self.is_vertical():
                        cr.rotate(-math.pi / 2)
                        cr.translate(-self.width, 0)
            else:
                if self.is_vertical():
                    cr.rotate(-math.pi / 2)
                    cr.translate(-self.width, 0)
        else:
            if self.port_category in ('sink', 'source', 'e_left', 'e_right', 'b_left', 'b_right'):
                if self.is_vertical():
                    cr.rotate(-math.pi / 2)
                    cr.translate(-self.width, 0)

        if self.port_category in ('e_top', 'e_bottom', 'b_top', 'b_bottom'):
            if self.is_horizontal():
                cr.rotate(-math.pi / 2)
                cr.translate(-self.width, 0)

        if self.parent_block.name.startswith('connector_f'):
            if self.is_source:
                index1 = self.port_index
                port_pos = self.parent_block.params['output' + str(int(index1)+1)].get_value()
            else:
                port_pos = 'left'
        else:
            _mirror = self.parent_block.mirror
            if _mirror == 'none':
                port_pos = {
                  'sink'    : 'left',
                  'source'  : 'right',
                  'e_left'  : 'left',
                  'e_right' : 'right',
                  'e_top'   : 'top',
                  'e_bottom': 'bottom',
                  'b_left'  : 'left',
                  'b_right' : 'right',
                  'b_top'   : 'top',
                  'b_bottom': 'bottom',
                }[self.port_category]
            elif _mirror == 'v':
                port_pos = {
                  'sink'    : 'right',
                  'source'  : 'left',
                  'e_left'  : 'right',
                  'e_right' : 'left',
                  'e_top'   : 'top',
                  'e_bottom': 'bottom',
                  'b_left'  : 'right',
                  'b_right' : 'left',
                  'b_top'   : 'top',
                  'b_bottom': 'bottom',
                }[self.port_category]
            elif _mirror == 'h':
                port_pos = {
                  'sink'    : 'left',
                  'source'  : 'right',
                  'e_left'  : 'left',
                  'e_right' : 'right',
                  'e_top'   : 'bottom',
                  'e_bottom': 'top',
                  'b_left'  : 'left',
                  'b_right' : 'right',
                  'b_top'   : 'bottom',
                  'b_bottom': 'top',
                }[self.port_category]

        _mirror = self.parent_block.mirror

        d1 = {
            ('top'   ,   0) : 'L',
            ('top'   ,  90) : 'R',
            ('top'   , 180) : 'R',
            ('top'   , 270) : 'L',
            ('bottom',   0) : 'R',
            ('bottom',  90) : 'L',
            ('bottom', 180) : 'L',
            ('bottom', 270) : 'R',
            ('left'  ,   0) : 'R',
            ('left'  ,  90) : 'R',
            ('left'  , 180) : 'L',
            ('left'  , 270) : 'L',
            ('right' ,   0) : 'L',
            ('right' ,  90) : 'L',
            ('right' , 180) : 'R',
            ('right' , 270) : 'R',
        }

        align = d1[(port_pos, self.rotation)]

        s_font = "Sans " + str(Constants.PORT_LABEL_FONTSIZE)
        self.label_layout.set_markup('<span font_desc="{font}">{name}</span>'.format(
            name=Utils.encode(self.name), font=s_font))

        if align == 'L':
            self.label_layout.set_alignment(Pango.Alignment.LEFT)
            self._label_layout_offsets = [5, -Constants.PORT_LABEL_OFFSET]
        elif align == 'R':
            self.label_layout.set_alignment(Pango.Alignment.RIGHT)
            self._label_layout_offsets = [-5, -Constants.PORT_LABEL_OFFSET]

        cr.translate(*self._label_layout_offsets)

        cr.set_source_rgba(*self._font_color)

        PangoCairo.update_layout(cr, self.label_layout)
        PangoCairo.show_layout(cr, self.label_layout)

    @property
    def connector_coordinate_absolute(self):
        """the coordinate where connections may attach to"""

        return [sum(c) for c in zip(
            self._connector_coordinate,   # relative to port
            self.coordinate,              # relative to block
            self.parent_block.coordinate  # abs
        )]

    @property
    def connector_direction(self):
        """Get the direction that the socket points: 0,90,180,270."""

        if self.parent_block.name.startswith('connector_f'):
            if self.is_source:
                index1 = self.port_index
                pos = self.parent_block.params['output' + str(int(index1)+1)].get_value()
                if pos == 'top':
                    return (self.rotation + 90) % 360
                elif pos == 'bottom':
                    return (self.rotation + 270) % 360
                elif pos == 'right':
                    return self.rotation
            else:
                return (self.rotation + 180) % 360
        else:
            _mirror = self.parent_block.mirror
            r1 = self.rotation

            if _mirror == 'none':
                if self.is_source or self.is_e_right or self.is_b_right:
                    r = r1
                elif self.is_sink or self.is_e_left or self.is_b_left:
                    r = (r1 + 180) % 360
                elif self.is_e_top or self.is_b_top:
                    r = (r1 + 90) % 360
                elif self.is_e_bottom or self.is_b_bottom:
                    r = (r1 + 270) % 360
            elif _mirror == 'v':
                if self.is_source or self.is_e_right or self.is_b_right:
                    r = (r1 + 180) % 360
                elif self.is_sink or self.is_e_left or self.is_b_left:
                    r = r1
                elif self.is_e_top or self.is_b_top:
                    r = (r1 + 90) % 360
                elif self.is_e_bottom or self.is_b_bottom:
                    r = (r1 + 270) % 360
            elif _mirror == 'h':
                if self.is_source or self.is_e_right or self.is_b_right:
                    r = r1
                elif self.is_sink or self.is_e_left or self.is_b_left:
                    r = (r1 + 180) % 360
                elif self.is_e_top or self.is_b_top:
                    r = (r1 + 270) % 360
                elif self.is_e_bottom or self.is_b_bottom:
                    r = (r1 + 90) % 360

            return r

    @nop_write
    @property
    def rotation(self):
        return self.parent_block.rotation

    def rotate(self, direction):
        """
        Rotate the parent rather than self.

        Args:
            direction: degrees to rotate
        """
        self.parent_block.rotate(direction)

    def move(self, delta_coor):
        """Move the parent rather than self."""
        self.parent_block.move(delta_coor)

    @property
    def highlighted(self):
        return self.parent_block.highlighted

    @highlighted.setter
    def highlighted(self, value):
        self.parent_block.highlighted = value

    @property
    def _show_label(self):
        """
        Figure out if the label should be hidden

        Returns:
            true if the label should not be shown
        """
#       if self._hovering:
#           print('self._hovering:', self._hovering)
#           print('self.force_show_label:', self.force_show_label)
#           print('Actions.TOGGLE_AUTO_HIDE_PORT_LABELS.get_active():',
#             Actions.TOGGLE_AUTO_HIDE_PORT_LABELS.get_active())
#       return self._hovering
#       return self._hovering or self.force_show_label
        return self._hovering or self.force_show_label or not Actions.TOGGLE_AUTO_HIDE_PORT_LABELS.get_active()

    def mouse_over(self):
        """
        Called from flow graph on mouse-over
        """
#       print('port.py: mouse_over:')
        changed = not self._show_label
        self._hovering = True
        return changed

    def mouse_out(self):
        """
        Called from flow graph on mouse-out
        """
        label_was_shown = self._show_label
        self._hovering = False
        return label_was_shown != self._show_label



# Copyright 2008-2016 Free Software Foundation, Inc.
# This file is part of GNU Radio
#
# GNU Radio Companion is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# GNU Radio Companion is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

from __future__ import absolute_import, print_function

from codecs import open
from collections import namedtuple
import os
import logging
from itertools import chain
import re

from . import (
    Messages, Constants,
    blocks, params, ports, errors, utils, schema_checker
)
from .utils import backports
from .utils import descriptors

from .utils import gutils as gu

from .Config import Config
from .base import Element
from .io import yaml
from .generator import Generator
from .FlowGraph import FlowGraph
from .Connection import Connection

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)

class Platform(Element):

    def __init__(self, *args, **kwargs):
        """ Make a platform for GNU Radio """
        Element.__init__(self, parent=None)

        self.config = self.Config(*args, **kwargs)

        self.blocks = self.block_classes
        self.domains = {}

        self.dict_outvars = {}

        self.d_slvparms = {}
        filename = Config.gseim_exec_dir + '/gseim_slvparms.in'
        gu.get_parms(filename, self.d_slvparms)
        self.d_slvparms['block_index'] = {'options': ['none'], 'default': '0', 'category': 'none'}

        self.d_slv_categories = {} 

        gu.prepare_dict_1(self.d_slvparms, self.d_slv_categories)

        self.d_outparms = {}
        filename = Config.gseim_exec_dir + '/gseim_outparms.in'
        gu.get_parms(filename, self.d_outparms)

        self._block_categories = {}
        self._auto_hier_block_generate_chain = set()

        if not yaml.__with_libyaml__:
            logger.warning("Slow YAML loading (libyaml not available)")

    def __str__(self):
        return 'Platform - {}'.format(self.config.name)

    @staticmethod
    def find_file_in_paths(filename, paths, cwd):
        """Checks the provided paths relative to cwd for a certain filename"""
        if not os.path.isdir(cwd):
            cwd = os.path.dirname(cwd)
        if isinstance(paths, str):
            paths = (p for p in paths.split(':') if p)

        for path in paths:
            path = os.path.expanduser(path)
            if not os.path.isabs(path):
                path = os.path.normpath(os.path.join(cwd, path))
            file_path = os.path.join(path, filename)
            if os.path.exists(os.path.normpath(file_path)):
                return file_path

    def load_and_generate_flow_graph(self, file_path, out_path=None, hier_only=False):
        """Loads a flow graph from file and generates it"""
        Messages.set_indent(len(self._auto_hier_block_generate_chain))
        Messages.send('>>> Loading: {}\n'.format(file_path))
        if file_path in self._auto_hier_block_generate_chain:
            Messages.send('    >>> Warning: cyclic hier_block dependency\n')
            return None, None

        self._auto_hier_block_generate_chain.add(file_path)
        try:
            flow_graph = self.make_flow_graph()
            flow_graph.grc_file_path = file_path
            # Other, nested hier_blocks might be auto-loaded here

            flow_graph.import_data(self.parse_flow_graph(file_path))

            flow_graph.rewrite()
            flow_graph.validate()
            if not flow_graph.is_valid():
                raise Exception('Flowgraph invalid')
            if hier_only and not flow_graph.get_option('generate_options').startswith('hb'):
                raise Exception('Not a hier block')
        except Exception as e:
            Messages.send('>>> Load Error: {}: {}\n'.format(file_path, str(e)))
            return None, None
        finally:
            self._auto_hier_block_generate_chain.discard(file_path)
            Messages.set_indent(len(self._auto_hier_block_generate_chain))

        try:
            print('grc/core/platform.py: out_path =', out_path, ' file_path =', file_path)
            generator = self.Generator(flow_graph, out_path or file_path)
            Messages.send('>>> Generating: {}\n'.format(generator.file_path))
            generator.write()
        except Exception as e:
            Messages.send('>>> platform.py: Generate Error: {}: {}\n'.format(file_path, str(e)))
            return None, None

        if flow_graph.get_option('generate_options').startswith('hb'):
            # self.load_block_xml(generator.file_path_xml)
            # TODO: implement yml output for hier blocks
            pass

        return flow_graph, 'dummy_string'

    def build_library(self, path=None):
        """load the blocks and block tree from the search paths

        path: a list of paths/files to search in or load (defaults to config)
        """

        # Reset
        self.blocks.clear()
        self.domains.clear()
        self._block_categories.clear()

        for file_path in self._iter_files_in_block_path(path):

            if file_path.endswith('block.yml'):
                loader = self.load_block_description
                scheme = schema_checker.BLOCK_SCHEME
            elif file_path.endswith('.tree.yml'):
                loader = self.load_category_tree_description
                scheme = None
            else:
                continue

            try:
                checker = schema_checker.Validator(scheme)
                with open(os.path.expanduser(file_path), encoding='utf-8') as fp:
                    data = yaml.safe_load(fp)
                passed = checker.run(data)
                for msg in checker.messages:
                    logger.warning('{:<40s} {}'.format(os.path.basename(file_path), msg))
                if not passed:
                    logger.info('YAML schema check failed for: ' + file_path)

                loader(data, file_path)
            except Exception as error:
                logger.exception('Error while loading %s', file_path)
                logger.exception(error)
                Messages.flowgraph_error = error
                Messages.flowgraph_error_file = file_path
                continue

        for key, block in iter(self.blocks.items()):
            category = self._block_categories.get(key, block.category)
            if not category:
                continue
            root = category[0]
            if root.startswith('[') and root.endswith(']'):
                category[0] = root[1:-1]
            else:
                category.insert(0, Constants.DEFAULT_BLOCK_MODULE_NAME)
            block.category = category

    def _iter_files_in_block_path(self, path=None, ext='yml'):
        """Iterator for block descriptions and category trees"""
        for entry in (path or self.config.block_paths):
            if os.path.isfile(entry):
                yield entry
            elif os.path.isdir(entry):
                for dirpath, dirnames, filenames in os.walk(entry):
                    for filename in sorted(filter(lambda f: f.endswith('.' + ext), filenames)):
                        yield os.path.join(dirpath, filename)
            else:
                logger.debug('Ignoring invalid path entry %r', entry)

    # Description File Loaders
    # region loaders
    def load_block_description(self, data, file_path):
        log = logger.getChild('block_loader')

        # don't load future block format versions

        block_id = data['id'] = data['id'].rstrip('_')
        print('load_block_description:', file_path)
        if 'outvars' in data.keys():
            block_name = file_path.split('/')[-1].split('.')[0]
            self.dict_outvars[block_name] = data['outvars']

        if block_id in self.block_classes_build_in:
            log.warning('Not overwriting build-in block %s with %s', block_id, file_path)
            return
        if block_id in self.blocks:
            log.warning('Block with id "%s" loaded from\n  %s\noverwritten by\n  %s',
                        block_id, self.blocks[block_id].loaded_from, file_path)

        try:
            block_cls = self.blocks[block_id] = self.new_block_class(**data)
            block_cls.loaded_from = file_path
        except errors.BlockLoadError as error:
            log.error('Unable to load block %s', block_id)
            log.exception(error)
            return

    def load_domain_description(self, data, file_path):
        log = logger.getChild('domain_loader')
        domain_id = data['id']
        if domain_id in self.domains:  # test against repeated keys
            log.debug('Domain "{}" already exists. Ignoring: %s', file_path)
            return

        color = data.get('color', '')
        if color.startswith('#'):
            try:
                tuple(int(color[o:o + 2], 16) / 255.0 for o in (1, 3, 5))
            except ValueError:
                log.warning('Cannot parse color code "%s" in %s', color, file_path)
                return

# todo
# check for e nodes
    def load_category_tree_description(self, data, file_path):
        """Parse category tree file and add it to list"""
        log = logger.getChild('tree_loader')
        log.debug('### Loading %s', file_path)
        path = []

        def load_category(name, elements):
            if not isinstance(name, str):
                log.debug('Invalid name %r', name)
                return
            path.append(name)
            for element in utils.to_list(elements):
                if isinstance(element, str):
                    block_id = element
                    self._block_categories[block_id] = list(path)
                elif isinstance(element, dict):
                    load_category(*next(iter(element.items())))
                else:
                    log.debug('Ignoring some elements of %s', name)
            path.pop()

        try:
            module_name, categories = next(iter(data.items()))
        except (AttributeError, StopIteration):
            log.warning('no valid data found')
        else:
            load_category(module_name, categories)

    # Access
    def parse_flow_graph(self, filename):
        """
        Parse a saved flow graph file.
        Ensure that the file exists, and passes the dtd check.

        Args:
            filename: the flow graph file

        Returns:
            nested data
        @throws exception if the validation fails
        """
        filename = filename or self.config.default_flow_graph
        with open(os.path.expanduser(filename), encoding='utf-8') as fp:

            fp.seek(0)

            data = yaml.safe_load(fp)
            validator = schema_checker.Validator(schema_checker.FLOW_GRAPH_SCHEME)
            validator.run(data)

        return data

    def save_flow_graph(self, filename, flow_graph):
        data = flow_graph.export_data()

        try:
            data['connections'] = [yaml.ListFlowing(i) for i in data['connections']]
        except KeyError:
            pass

        if flow_graph.get_option('generate_options').startswith('hb'):
            if 'drawing_scheme' not in data['gparms'].keys():
                data['gparms']['drawing_scheme'] = 'name'
            if 'rotate_strict' not in data['gparms'].keys():
                data['gparms']['rotate_strict'] = 'no'
            if 'mirror' not in data['gparms'].keys():
                data['gparms']['mirror'] = 'none'
            if 'port_sep_x' not in data['gparms'].keys():
                data['gparms']['port_sep_x'] = '4'
            if 'port_sep_y' not in data['gparms'].keys():
                data['gparms']['port_sep_y'] = '4'
            if 'port_block_x' not in data['gparms'].keys():
                data['gparms']['port_block_x'] = '4'
            if 'port_block_y' not in data['gparms'].keys():
                data['gparms']['port_block_y'] = '4'
            if 'port_offset_l' not in data['gparms'].keys():
                data['gparms']['port_offset_l'] = '0'
            if 'port_offset_r' not in data['gparms'].keys():
                data['gparms']['port_offset_r'] = '0'
            if 'port_offset_t' not in data['gparms'].keys():
                data['gparms']['port_offset_t'] = '0'
            if 'port_offset_b' not in data['gparms'].keys():
                data['gparms']['port_offset_b'] = '0'

        try:
            for d in chain([data['options']], data['blocks']):
                d['states']['coordinate'] = yaml.ListFlowing(d['states']['coordinate'])
        except KeyError:
            pass

        out = yaml.dump(data, indent=2)

        with open(os.path.expanduser(filename), 'w', encoding='utf-8') as fp:
            fp.write(out)

    def get_generate_options(self):
        for param in self.block_classes['options'].parameters_data:
            if param.get('id') == 'generate_options':
                break
        else:
            return []
        generate_mode_default = param.get('default')
        return [(value, name, value == generate_mode_default)
                for value, name in zip(param['options'], param['option_labels'])]

    # Factories
    Config = Config
    Domain = namedtuple('Domain', 'name multi_in multi_out color')
    Generator = Generator
    FlowGraph = FlowGraph
    Connection = Connection

    block_classes_build_in = blocks.build_ins
    block_classes = utils.backports.ChainMap({}, block_classes_build_in)  # separates build-in from loaded blocks)

    port_classes = {
        None: ports.Port,  # default
    }
    param_classes = {
        None: params.Param,  # default
    }

    def make_flow_graph(self, from_filename=None):
        fg = self.FlowGraph(parent=self)
        if from_filename:
            data = self.parse_flow_graph(from_filename)
            fg.grc_file_path = from_filename
            fg.import_data(data)
        return fg

    def new_block_class(self, **data):
        return blocks.build(**data)

    def make_block(self, parent, block_id, **kwargs):
        cls = self.block_classes[block_id]
        return cls(parent, **kwargs)

    def make_param(self, parent, **kwargs):
        cls = self.param_classes[kwargs.pop('cls_key', None)]
        return cls(parent, **kwargs)

    def make_port(self, parent, **kwargs):
        cls = self.port_classes[kwargs.pop('cls_key', None)]
        return cls(parent, **kwargs)
